import decimal
import enum
import json
import unittest
import uuid

from django.contrib.admin.utils import display_for_field
from django.core import checks, exceptions, serializers, validators
from django.core.exceptions import FieldError
from django.core.management import call_command
from django.db import IntegrityError, connection, models
from django.db.models.expressions import Exists, OuterRef, Value
from django.db.models.functions import Upper
from django.test import (
    SimpleTestCase,
    TestCase,
    TransactionTestCase,
    override_settings,
)
from django.test.utils import isolate_apps
from django.utils import timezone

from django_mongodb_backend.fields import ArrayField

from .models import (
    ArrayEnumModel,
    ArrayFieldSubclass,
    CharArrayModel,
    DateTimeArrayModel,
    IntegerArrayModel,
    NestedIntegerArrayModel,
    NullableIntegerArrayModel,
    OtherTypesArrayModel,
    Tag,
)


@isolate_apps("model_fields_")
class BasicTests(SimpleTestCase):
    def test_get_field_display(self):
        class MyModel(models.Model):
            field = ArrayField(
                models.CharField(max_length=16),
                choices=[
                    ["Media", [(["vinyl", "cd"], "Audio")]],
                    (("mp3", "mp4"), "Digital"),
                ],
            )

        tests = (
            (["vinyl", "cd"], "Audio"),
            (("mp3", "mp4"), "Digital"),
            (("a", "b"), "('a', 'b')"),
            (["c", "d"], "['c', 'd']"),
        )
        for value, display in tests:
            with self.subTest(value=value, display=display):
                instance = MyModel(field=value)
                self.assertEqual(instance.get_field_display(), display)

    def test_get_field_display_nested_array(self):
        class MyModel(models.Model):
            field = ArrayField(
                ArrayField(models.CharField(max_length=16)),
                choices=[
                    [
                        "Media",
                        [([["vinyl", "cd"], ("x",)], "Audio")],
                    ],
                    ((["mp3"], ("mp4",)), "Digital"),
                ],
            )

        tests = (
            ([["vinyl", "cd"], ("x",)], "Audio"),
            ((["mp3"], ("mp4",)), "Digital"),
            ((("a", "b"), ("c",)), "(('a', 'b'), ('c',))"),
            ([["a", "b"], ["c"]], "[['a', 'b'], ['c']]"),
        )
        for value, display in tests:
            with self.subTest(value=value, display=display):
                instance = MyModel(field=value)
                self.assertEqual(instance.get_field_display(), display)

    def test_deconstruct(self):
        field = ArrayField(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(kwargs.keys(), {"base_field"})
        new = ArrayField(*args, **kwargs)
        self.assertEqual(type(new.base_field), type(field.base_field))
        self.assertIsNot(new.base_field, field.base_field)

    def test_deconstruct_with_max_size(self):
        field = ArrayField(models.IntegerField(), max_size=3)
        name, path, args, kwargs = field.deconstruct()
        new = ArrayField(*args, **kwargs)
        self.assertEqual(new.max_size, field.max_size)

    def test_deconstruct_args(self):
        field = ArrayField(models.CharField(max_length=20))
        name, path, args, kwargs = field.deconstruct()
        new = ArrayField(*args, **kwargs)
        self.assertEqual(new.base_field.max_length, field.base_field.max_length)

    def test_subclass_deconstruct(self):
        field = ArrayField(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "django_mongodb_backend.fields.ArrayField")

        field = ArrayFieldSubclass()
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "model_fields_.models.ArrayFieldSubclass")


class SaveLoadTests(TestCase):
    def test_integer(self):
        instance = IntegerArrayModel(field=[1, 2, 3])
        instance.save()
        loaded = IntegerArrayModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_char(self):
        instance = CharArrayModel(field=["hello", "goodbye"])
        instance.save()
        loaded = CharArrayModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_dates(self):
        instance = DateTimeArrayModel(
            datetimes=[timezone.now().replace(microsecond=0)],
            dates=[timezone.now().date()],
            times=[timezone.now().time().replace(microsecond=0)],
        )
        instance.save()
        loaded = DateTimeArrayModel.objects.get()
        self.assertEqual(instance.datetimes, loaded.datetimes)
        self.assertEqual(instance.dates, loaded.dates)
        self.assertEqual(instance.times, loaded.times)

    def test_tuples(self):
        instance = IntegerArrayModel(field=(1,))
        instance.save()
        loaded = IntegerArrayModel.objects.get()
        self.assertSequenceEqual(instance.field, loaded.field)

    def test_integers_passed_as_strings(self):
        # This checks that get_prep_value() is deferred properly.
        instance = IntegerArrayModel(field=["1"])
        instance.save()
        loaded = IntegerArrayModel.objects.get()
        self.assertEqual(loaded.field, [1])

    def test_default_null(self):
        instance = NullableIntegerArrayModel()
        instance.save()
        loaded = NullableIntegerArrayModel.objects.get(pk=instance.pk)
        self.assertIsNone(loaded.field)
        self.assertEqual(instance.field, loaded.field)

    def test_null_handling(self):
        instance = NullableIntegerArrayModel(field=None)
        instance.save()
        loaded = NullableIntegerArrayModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_save_null_in_non_null(self):
        instance = IntegerArrayModel(field=None)
        msg = "You can't set field (a non-nullable field) to None."
        with self.assertRaisesMessage(IntegrityError, msg):
            instance.save()

    def test_nested(self):
        instance = NestedIntegerArrayModel(field=[[1, 2], [3, 4]])
        instance.save()
        loaded = NestedIntegerArrayModel.objects.get()
        self.assertEqual(instance.field, loaded.field)

    def test_other_array_types(self):
        instance = OtherTypesArrayModel(
            ips=["192.168.0.1", "::1"],
            uuids=[uuid.uuid4()],
            decimals=[decimal.Decimal(1.25), 1.75],
            tags=[Tag(1), Tag(2), Tag(3)],
            json=[{"a": 1}, {"b": 2}],
        )
        instance.save()
        loaded = OtherTypesArrayModel.objects.get()
        self.assertEqual(instance.ips, loaded.ips)
        self.assertEqual(instance.uuids, loaded.uuids)
        self.assertEqual(instance.decimals, loaded.decimals)
        self.assertEqual(instance.tags, loaded.tags)
        self.assertEqual(instance.json, loaded.json)

    def test_null_from_db_value_handling(self):
        instance = OtherTypesArrayModel.objects.create(
            ips=["192.168.0.1", "::1"],
            uuids=[uuid.uuid4()],
            decimals=[decimal.Decimal(1.25), 1.75],
            tags=None,
        )
        instance.refresh_from_db()
        self.assertIsNone(instance.tags)
        self.assertEqual(instance.json, [])

    def test_model_set_on_base_field(self):
        instance = IntegerArrayModel()
        field = instance._meta.get_field("field")
        self.assertEqual(field.model, IntegerArrayModel)
        self.assertEqual(field.base_field.model, IntegerArrayModel)

    def test_nested_nullable_base_field(self):
        instance = NullableIntegerArrayModel.objects.create(
            field_nested=[[None, None], [None, None]],
        )
        self.assertEqual(instance.field_nested, [[None, None], [None, None]])


class QueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.objs = NullableIntegerArrayModel.objects.bulk_create(
            [
                NullableIntegerArrayModel(order=1, field=[1]),
                NullableIntegerArrayModel(order=2, field=[2]),
                NullableIntegerArrayModel(order=3, field=[2, 3]),
                NullableIntegerArrayModel(order=4, field=[20, 30, 40]),
                NullableIntegerArrayModel(order=5, field=None),
            ]
        )

    def test_empty_list(self):
        NullableIntegerArrayModel.objects.create(field=[])
        obj = (
            NullableIntegerArrayModel.objects.annotate(
                empty_array=models.Value([], output_field=ArrayField(models.IntegerField())),
            )
            .filter(field=models.F("empty_array"))
            .get()
        )
        self.assertEqual(obj.field, [])
        self.assertEqual(obj.empty_array, [])

    def test_exact(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__exact=[1]), self.objs[:1]
        )

    def test_exact_null_only_array(self):
        obj = NullableIntegerArrayModel.objects.create(field=[None], field_nested=[None, None])
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__exact=[None]), [obj]
        )
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field_nested__exact=[None, None]),
            [obj],
        )

    def test_exact_null_only_nested_array(self):
        obj1 = NullableIntegerArrayModel.objects.create(field_nested=[[None, None]])
        obj2 = NullableIntegerArrayModel.objects.create(
            field_nested=[[None, None], [None, None]],
        )
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(
                field_nested__exact=[[None, None]],
            ),
            [obj1],
        )
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(
                field_nested__exact=[[None, None], [None, None]],
            ),
            [obj2],
        )

    def test_exact_with_expression(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__exact=[Value(1)]),
            self.objs[:1],
        )

    def test_exact_charfield(self):
        instance = CharArrayModel.objects.create(field=["text"])
        self.assertSequenceEqual(CharArrayModel.objects.filter(field=["text"]), [instance])

    def test_exact_nested(self):
        instance = NestedIntegerArrayModel.objects.create(field=[[1, 2], [3, 4]])
        self.assertSequenceEqual(
            NestedIntegerArrayModel.objects.filter(field=[[1, 2], [3, 4]]), [instance]
        )

    def test_isnull(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__isnull=True), self.objs[-1:]
        )

    def test_gt(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__gt=[0]), self.objs[:4]
        )

    def test_lt(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__lt=[2]), self.objs[:1]
        )

    def test_in(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__in=[[1], [2]]),
            self.objs[:2],
        )

    def test_in_subquery(self):
        IntegerArrayModel.objects.create(field=[2, 3])
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(
                field__in=IntegerArrayModel.objects.values_list("field", flat=True)
            ),
            self.objs[2:3],
        )

    @unittest.expectedFailure
    def test_in_including_F_object(self):
        # Array objects passed to filters can be constructed to contain
        # F objects. This doesn't work on PostgreSQL either (#27095).
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__in=[[models.F("id")]]),
            self.objs[:2],
        )

    def test_in_as_F_object(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__in=[models.F("field")]),
            # Unlike PostgreSQL, MongoDB returns documents with field=null,
            # i.e. null is in [null]. It seems okay to leave this alone rather
            # than filtering out null in all $in queries. Feel free to
            # reconsider this decision if the behavior is problematic in some
            # other query.
            self.objs,
        )

    def test_contained_by(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__contained_by=[1, 2]),
            self.objs[:2],
        )

    def test_contained_by_including_F_object(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__contained_by=[models.F("order"), 2]),
            self.objs[:3],
        )

    def test_contains(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__contains=[2]),
            self.objs[1:3],
        )

    def test_contains_subquery(self):
        IntegerArrayModel.objects.create(field=[2, 3])
        inner_qs = IntegerArrayModel.objects.values_list("field", flat=True)
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__contains=inner_qs[:1]),
            self.objs[2:3],
        )
        inner_qs = IntegerArrayModel.objects.filter(field__contains=OuterRef("field"))
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(Exists(inner_qs)),
            self.objs[1:3],
        )

    def test_contained_by_subquery(self):
        IntegerArrayModel.objects.create(field=[2, 3])
        inner_qs = IntegerArrayModel.objects.values_list("field", flat=True)
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__contained_by=inner_qs[:1]),
            self.objs[1:3],
        )
        IntegerArrayModel.objects.create(field=[2])
        inner_qs = IntegerArrayModel.objects.filter(field__contained_by=OuterRef("field"))
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(Exists(inner_qs)),
            self.objs[1:3],
        )

    def test_contains_including_expression(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(
                field__contains=[2, Value(6) / Value(2)],
            ),
            self.objs[2:3],
        )

    def test_icontains(self):
        instance = CharArrayModel.objects.create(field=["FoO"])
        self.assertSequenceEqual(CharArrayModel.objects.filter(field__icontains="foo"), [instance])

    def test_contains_charfield(self):
        self.assertSequenceEqual(CharArrayModel.objects.filter(field__contains=["text"]), [])

    def test_contained_by_charfield(self):
        self.assertSequenceEqual(CharArrayModel.objects.filter(field__contained_by=["text"]), [])

    def test_overlap_charfield(self):
        self.assertSequenceEqual(CharArrayModel.objects.filter(field__overlap=["text"]), [])

    def test_overlap_charfield_including_expression(self):
        obj_1 = CharArrayModel.objects.create(field=["TEXT", "lower text"])
        obj_2 = CharArrayModel.objects.create(field=["lower text", "TEXT"])
        CharArrayModel.objects.create(field=["lower text", "text"])
        self.assertSequenceEqual(
            CharArrayModel.objects.filter(
                field__overlap=[
                    Upper(Value("text")),
                    "other",
                ]
            ),
            [obj_1, obj_2],
        )

    def test_overlap_values(self):
        qs = NullableIntegerArrayModel.objects.filter(order__lt=3)
        self.assertCountEqual(
            NullableIntegerArrayModel.objects.filter(
                field__overlap=qs.values_list("field"),
            ),
            self.objs[:3],
        )
        self.assertCountEqual(
            NullableIntegerArrayModel.objects.filter(
                field__overlap=qs.values("field"),
            ),
            self.objs[:3],
        )

    def test_index(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__0=2), self.objs[1:3]
        )

    def test_index_chained(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__0__lt=3), self.objs[0:3]
        )

    def test_index_nested(self):
        instance = NestedIntegerArrayModel.objects.create(field=[[1, 2], [3, 4]])
        self.assertSequenceEqual(NestedIntegerArrayModel.objects.filter(field__0__0=1), [instance])

    def test_index_used_on_nested_data(self):
        instance = NestedIntegerArrayModel.objects.create(field=[[1, 2], [3, 4]])
        self.assertSequenceEqual(
            NestedIntegerArrayModel.objects.filter(field__0=[1, 2]), [instance]
        )

    def test_overlap(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__overlap=[1, 2]),
            self.objs[0:3],
        )

    def test_index_annotation(self):
        qs = NullableIntegerArrayModel.objects.annotate(second=models.F("field__1"))
        self.assertCountEqual(
            qs.values_list("second", flat=True),
            [None, None, None, 3, 30],
        )

    def test_len(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__len__lte=2), self.objs[0:3]
        )

    def test_len_empty_array(self):
        obj = NullableIntegerArrayModel.objects.create(field=[])
        self.assertSequenceEqual(NullableIntegerArrayModel.objects.filter(field__len=0), [obj])

    def test_slice(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__0_1=[2]), self.objs[1:3]
        )

        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(field__0_2=[2, 3]), self.objs[2:3]
        )

    def test_order_by_index(self):
        more_objs = (
            NullableIntegerArrayModel.objects.create(field=[1, 637]),
            NullableIntegerArrayModel.objects.create(field=[2, 1]),
            NullableIntegerArrayModel.objects.create(field=[3, -98123]),
            NullableIntegerArrayModel.objects.create(field=[4, 2]),
        )
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.order_by("field__1"),
            [
                self.objs[0],
                self.objs[1],
                self.objs[4],
                more_objs[2],
                more_objs[1],
                more_objs[3],
                self.objs[2],
                self.objs[3],
                more_objs[0],
            ],
        )

    def test_slice_nested(self):
        instance = NestedIntegerArrayModel.objects.create(field=[[1, 2], [3, 4]])
        self.assertSequenceEqual(
            NestedIntegerArrayModel.objects.filter(field__0__0_1=[1]), [instance]
        )

    def test_slice_annotation(self):
        qs = NullableIntegerArrayModel.objects.annotate(
            first_two=models.F("field__0_2"),
        )
        self.assertCountEqual(
            qs.values_list("first_two", flat=True),
            [None, [1], [2], [2, 3], [20, 30]],
        )

    def test_usage_in_subquery(self):
        self.assertSequenceEqual(
            NullableIntegerArrayModel.objects.filter(
                id__in=NullableIntegerArrayModel.objects.filter(field__len=3)
            ),
            [self.objs[3]],
        )

    def test_enum_lookup(self):
        class TestEnum(enum.Enum):
            VALUE_1 = "value_1"

        instance = ArrayEnumModel.objects.create(array_of_enums=[TestEnum.VALUE_1])
        self.assertSequenceEqual(
            ArrayEnumModel.objects.filter(array_of_enums__contains=[TestEnum.VALUE_1]),
            [instance],
        )

    def test_unsupported_lookup(self):
        msg = "Unsupported lookup '0_bar' for ArrayField or join on the field not " "permitted."
        with self.assertRaisesMessage(FieldError, msg):
            list(NullableIntegerArrayModel.objects.filter(field__0_bar=[2]))

        msg = "Unsupported lookup '0bar' for ArrayField or join on the field not " "permitted."
        with self.assertRaisesMessage(FieldError, msg):
            list(NullableIntegerArrayModel.objects.filter(field__0bar=[2]))


class DateTimeExactQueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        now = timezone.now()
        cls.datetimes = [now]
        cls.dates = [now.date()]
        cls.times = [now.time()]
        cls.objs = [
            DateTimeArrayModel.objects.create(
                datetimes=cls.datetimes, dates=cls.dates, times=cls.times
            ),
        ]

    def test_exact_datetimes(self):
        self.assertSequenceEqual(
            DateTimeArrayModel.objects.filter(datetimes=self.datetimes), self.objs
        )

    def test_exact_dates(self):
        self.assertSequenceEqual(DateTimeArrayModel.objects.filter(dates=self.dates), self.objs)

    def test_exact_times(self):
        self.assertSequenceEqual(DateTimeArrayModel.objects.filter(times=self.times), self.objs)


class OtherTypesExactQueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.ips = ["192.168.0.1", "::1"]
        cls.uuids = [uuid.uuid4()]
        cls.decimals = [decimal.Decimal(1.25), 1.75]
        cls.tags = [Tag(1), Tag(2), Tag(3)]
        cls.objs = [
            OtherTypesArrayModel.objects.create(
                ips=cls.ips,
                uuids=cls.uuids,
                decimals=cls.decimals,
                tags=cls.tags,
            )
        ]

    def test_exact_ip_addresses(self):
        self.assertSequenceEqual(OtherTypesArrayModel.objects.filter(ips=self.ips), self.objs)

    def test_exact_uuids(self):
        self.assertSequenceEqual(OtherTypesArrayModel.objects.filter(uuids=self.uuids), self.objs)

    def test_exact_decimals(self):
        self.assertSequenceEqual(
            OtherTypesArrayModel.objects.filter(decimals=self.decimals), self.objs
        )

    def test_exact_tags(self):
        self.assertSequenceEqual(OtherTypesArrayModel.objects.filter(tags=self.tags), self.objs)


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_base_field_errors(self):
        class MyModel(models.Model):
            field = ArrayField(models.CharField(max_length=-1))

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        # The inner CharField has a non-positive max_length.
        self.assertEqual(errors[0].id, "django_mongodb_backend.array.E001")
        msg = errors[0].msg
        self.assertIn("Base field for array has errors:", msg)
        self.assertIn("'max_length' must be a positive integer. (fields.E121)", msg)

    def test_base_field_warnings(self):
        class WarningField(models.IntegerField):
            def check(self):
                return [checks.Warning("Test warning", obj=self, id="test.E001")]

        class MyModel(models.Model):
            field = ArrayField(WarningField(), default=None)

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "django_mongodb_backend.array.W004")
        msg = errors[0].msg
        self.assertIn("Base field for array has warnings:", msg)
        self.assertIn("Test warning (test.E001)", msg)

    def test_invalid_base_fields(self):
        class MyModel(models.Model):
            field = ArrayField(models.ManyToManyField("model_fields_.IntegerArrayModel"))

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "django_mongodb_backend.array.E002")

    def test_both_size_and_max_size(self):
        class MyModel(models.Model):
            field = ArrayField(models.CharField(max_length=3), size=3, max_size=4)

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "django_mongodb_backend.array.E003")
        self.assertEqual(errors[0].msg, "ArrayField cannot have both size and max_size.")

    def test_invalid_default(self):
        class MyModel(models.Model):
            field = ArrayField(models.IntegerField(), default=[])

        model = MyModel()
        self.assertEqual(
            model.check(),
            [
                checks.Warning(
                    msg=(
                        "ArrayField default should be a callable instead of an "
                        "instance so that it's not shared between all field "
                        "instances."
                    ),
                    hint="Use a callable instead, e.g., use `list` instead of `[]`.",
                    obj=MyModel._meta.get_field("field"),
                    id="fields.E010",
                )
            ],
        )

    def test_valid_default(self):
        class MyModel(models.Model):
            field = ArrayField(models.IntegerField(), default=list)

        model = MyModel()
        self.assertEqual(model.check(), [])

    def test_valid_default_none(self):
        class MyModel(models.Model):
            field = ArrayField(models.IntegerField(), default=None)

        model = MyModel()
        self.assertEqual(model.check(), [])

    def test_nested_field_checks(self):
        """
        Nested ArrayFields are permitted.
        """

        class MyModel(models.Model):
            field = ArrayField(ArrayField(models.CharField(max_length=-1)))

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        # The inner CharField has a non-positive max_length.
        self.assertEqual(errors[0].id, "django_mongodb_backend.array.E001")
        self.assertIn("max_length", errors[0].msg)

    def test_choices_tuple_list(self):
        class MyModel(models.Model):
            field = ArrayField(
                models.CharField(max_length=16),
                choices=[
                    [
                        "Media",
                        [(["vinyl", "cd"], "Audio"), (("vhs", "dvd"), "Video")],
                    ],
                    (["mp3", "mp4"], "Digital"),
                ],
            )

        self.assertEqual(MyModel._meta.get_field("field").check(), [])


@isolate_apps("model_fields_")
class MigrationsTests(TransactionTestCase):
    available_apps = ["model_fields_"]

    @override_settings(
        MIGRATION_MODULES={
            "model_fields_": "model_fields_.array_default_migrations",
        }
    )
    def test_adding_field_with_default(self):
        class IntegerArrayDefaultModel(models.Model):
            field = ArrayField(models.IntegerField())

        table_name = "model_fields__integerarraydefaultmodel"
        self.assertNotIn(table_name, connection.introspection.table_names(None))
        # Create collection
        call_command("migrate", "model_fields_", "0001", verbosity=0)
        self.assertIn(table_name, connection.introspection.table_names(None))
        obj = IntegerArrayDefaultModel.objects.create(field=[1, 2, 3])
        # Add `field2 to IntegerArrayDefaultModel.
        call_command("migrate", "model_fields_", "0002", verbosity=0)

        class UpdatedIntegerArrayDefaultModel(models.Model):
            field = ArrayField(models.IntegerField())
            field_2 = ArrayField(models.IntegerField(), default=[])

            class Meta:
                db_table = "model_fields__integerarraydefaultmodel"

        obj = UpdatedIntegerArrayDefaultModel.objects.get()
        # The default is populated on existing documents.
        self.assertEqual(obj.field_2, [])
        # Cleanup.
        call_command("migrate", "model_fields_", "zero", verbosity=0)
        self.assertNotIn(table_name, connection.introspection.table_names(None))

    @override_settings(
        MIGRATION_MODULES={
            "model_fields_": "model_fields_.array_index_migrations",
        }
    )
    def test_adding_arrayfield_with_index(self):
        table_name = "model_fields__chartextarrayindexmodel"
        call_command("migrate", "model_fields_", verbosity=0)
        # All fields should have indexes.
        indexes = [
            c["columns"][0]
            for c in connection.introspection.get_constraints(None, table_name).values()
            if c["index"] and len(c["columns"]) == 1
        ]
        self.assertIn("char", indexes)
        self.assertIn("char2", indexes)
        self.assertIn("text", indexes)
        call_command("migrate", "model_fields_", "zero", verbosity=0)
        self.assertNotIn(table_name, connection.introspection.table_names(None))


class SerializationTests(SimpleTestCase):
    test_data = (
        '[{"fields": {"field": "[\\"1\\", \\"2\\", null]"}, '
        '"model": "model_fields_.integerarraymodel", "pk": null}]'
    )

    def test_dumping(self):
        instance = IntegerArrayModel(field=[1, 2, None])
        data = serializers.serialize("json", [instance])
        self.assertEqual(json.loads(data), json.loads(self.test_data))

    def test_loading(self):
        instance = next(serializers.deserialize("json", self.test_data)).object
        self.assertEqual(instance.field, [1, 2, None])


class ValidationTests(SimpleTestCase):
    def test_unbounded(self):
        field = ArrayField(models.IntegerField())
        msg = "Item 2 in the array did not validate: This field cannot be null."
        with self.assertRaisesMessage(exceptions.ValidationError, msg) as cm:
            field.clean([1, None], None)
        self.assertEqual(cm.exception.code, "item_invalid")

    def test_blank_true(self):
        field = ArrayField(models.IntegerField(blank=True, null=True))
        # This should not raise a validation error
        field.clean([1, None], None)

    def test_with_max_size(self):
        field = ArrayField(models.IntegerField(), max_size=3)
        field.clean([1, 2, 3], None)
        msg = "List contains 4 items, it should contain no more than 3."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean([1, 2, 3, 4], None)

    def test_with_max_size_singular(self):
        field = ArrayField(models.IntegerField(), max_size=1)
        field.clean([1], None)
        msg = "List contains 2 items, it should contain no more than 1."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean([1, 2], None)

    def test_with_size(self):
        field = ArrayField(models.IntegerField(), size=3)
        field.clean([1, 2, 3], None)
        msg = "List contains 4 items, it should contain 3."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean([1, 2, 3, 4], None)

    def test_with_size_singular(self):
        field = ArrayField(models.IntegerField(), size=2)
        field.clean([1, 2], None)
        msg = "List contains 1 item, it should contain 2."
        with self.assertRaisesMessage(exceptions.ValidationError, msg):
            field.clean([1], None)

    def test_nested_array_mismatch(self):
        field = ArrayField(ArrayField(models.IntegerField()))
        field.clean([[1, 2], [3, 4]], None)
        msg = "Nested arrays must have the same length."
        with self.assertRaisesMessage(exceptions.ValidationError, msg) as cm:
            field.clean([[1, 2], [3, 4, 5]], None)
        self.assertEqual(cm.exception.code, "nested_array_mismatch")

    def test_with_base_field_error_params(self):
        field = ArrayField(models.CharField(max_length=2))
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean(["abc"], None)
        self.assertEqual(len(cm.exception.error_list), 1)
        exception = cm.exception.error_list[0]
        self.assertEqual(
            exception.message,
            "Item 1 in the array did not validate: Ensure this value has at most 2 "
            "characters (it has 3).",
        )
        self.assertEqual(exception.code, "item_invalid")
        self.assertEqual(
            exception.params,
            {"nth": 1, "value": "abc", "limit_value": 2, "show_value": 3},
        )

    def test_with_validators(self):
        field = ArrayField(models.IntegerField(validators=[validators.MinValueValidator(1)]))
        field.clean([1, 2], None)
        with self.assertRaises(exceptions.ValidationError) as cm:
            field.clean([0], None)
        self.assertEqual(len(cm.exception.error_list), 1)
        exception = cm.exception.error_list[0]
        self.assertEqual(
            exception.message,
            "Item 1 in the array did not validate: Ensure this value is greater than "
            "or equal to 1.",
        )
        self.assertEqual(exception.code, "item_invalid")
        self.assertEqual(
            exception.params, {"nth": 1, "value": 0, "limit_value": 1, "show_value": 0}
        )


class AdminUtilsTests(SimpleTestCase):
    empty_value = "-empty-"

    def test_array_display_for_field(self):
        array_field = ArrayField(models.IntegerField())
        display_value = display_for_field([1, 2], array_field, self.empty_value)
        self.assertEqual(display_value, "1, 2")

    def test_array_with_choices_display_for_field(self):
        array_field = ArrayField(
            models.IntegerField(),
            choices=[
                ([1, 2, 3], "1st choice"),
                ([1, 2], "2nd choice"),
            ],
        )
        display_value = display_for_field([1, 2], array_field, self.empty_value)
        self.assertEqual(display_value, "2nd choice")
        display_value = display_for_field([99, 99], array_field, self.empty_value)
        self.assertEqual(display_value, self.empty_value)



from bson import ObjectId
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from django_mongodb_backend.forms.fields import ObjectIdField


class ObjectIdFieldTests(SimpleTestCase):
    def test_clean(self):
        field = ObjectIdField()
        value = field.clean("675747ec45260945758d76bc")
        self.assertEqual(value, ObjectId("675747ec45260945758d76bc"))

    def test_clean_objectid(self):
        field = ObjectIdField()
        value = field.clean(ObjectId("675747ec45260945758d76bc"))
        self.assertEqual(value, ObjectId("675747ec45260945758d76bc"))

    def test_clean_empty_string(self):
        field = ObjectIdField(required=False)
        value = field.clean("")
        self.assertEqual(value, None)

    def test_clean_invalid(self):
        field = ObjectIdField()
        with self.assertRaisesMessage(ValidationError, "Enter a valid Object Id."):
            field.clean("invalid")

    def test_prepare_value(self):
        field = ObjectIdField()
        value = field.prepare_value(ObjectId("675747ec45260945758d76bc"))
        self.assertEqual(value, "675747ec45260945758d76bc")



from .array import ArrayField
from .auto import ObjectIdAutoField
from .duration import register_duration_field
from .embedded_model import EmbeddedModelField
from .json import register_json_field
from .objectid import ObjectIdField

__all__ = [
    "register_fields",
    "ArrayField",
    "EmbeddedModelField",
    "ObjectIdAutoField",
    "ObjectIdField",
]


def register_fields():
    register_duration_field()
    register_json_field()



import json

from django.core import checks, exceptions
from django.db.models import DecimalField, Field, Func, IntegerField, Transform, Value
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.db.models.lookups import Exact, FieldGetDbPrepValueMixin, In, Lookup
from django.utils.translation import gettext_lazy as _

from ..forms import SimpleArrayField
from ..query_utils import process_lhs, process_rhs
from ..utils import prefix_validation_error
from ..validators import ArrayMaxLengthValidator, LengthValidator

__all__ = ["ArrayField"]


class AttributeSetter:
    def __init__(self, name, value):
        setattr(self, name, value)


class ArrayField(CheckFieldDefaultMixin, Field):
    empty_strings_allowed = False
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
        "nested_array_mismatch": _("Nested arrays must have the same length."),
    }
    _default_hint = ("list", "[]")

    def __init__(self, base_field, max_size=None, size=None, **kwargs):
        self.base_field = base_field
        self.max_size = max_size
        self.size = size
        if self.max_size:
            self.default_validators = [
                *self.default_validators,
                ArrayMaxLengthValidator(self.max_size),
            ]
        if self.size:
            self.default_validators = [
                *self.default_validators,
                LengthValidator(self.size),
            ]
        # For performance, only add a from_db_value() method if the base field
        # implements it.
        if hasattr(self.base_field, "from_db_value"):
            self.from_db_value = self._from_db_value
        super().__init__(**kwargs)

    @property
    def model(self):
        try:
            return self.__dict__["model"]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute 'model'" % self.__class__.__name__
            ) from None

    @model.setter
    def model(self, model):
        self.__dict__["model"] = model
        self.base_field.model = model

    @classmethod
    def _choices_is_value(cls, value):
        return isinstance(value, list | tuple) or super()._choices_is_value(value)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.base_field.remote_field:
            errors.append(
                checks.Error(
                    "Base field for array cannot be a related field.",
                    obj=self,
                    id="django_mongodb_backend.array.E002",
                )
            )
        else:
            base_checks = self.base_field.check()
            if base_checks:
                error_messages = "\n    ".join(
                    f"{base_check.msg} ({base_check.id})"
                    for base_check in base_checks
                    if isinstance(base_check, checks.Error)
                )
                if error_messages:
                    errors.append(
                        checks.Error(
                            f"Base field for array has errors:\n    {error_messages}",
                            obj=self,
                            id="django_mongodb_backend.array.E001",
                        )
                    )
                warning_messages = "\n    ".join(
                    f"{base_check.msg} ({base_check.id})"
                    for base_check in base_checks
                    if isinstance(base_check, checks.Warning)
                )
                if warning_messages:
                    errors.append(
                        checks.Warning(
                            f"Base field for array has warnings:\n    {warning_messages}",
                            obj=self,
                            id="django_mongodb_backend.array.W004",
                        )
                    )
        if self.size and self.max_size:
            errors.append(
                checks.Error(
                    "ArrayField cannot have both size and max_size.",
                    obj=self,
                    id="django_mongodb_backend.array.E003",
                )
            )
        return errors

    def set_attributes_from_name(self, name):
        super().set_attributes_from_name(name)
        self.base_field.set_attributes_from_name(name)

    @property
    def description(self):
        return f"Array of {self.base_field.description}"

    def db_type(self, connection):
        return "array"

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, list | tuple):
            # Workaround for https://code.djangoproject.com/ticket/35982
            # (fixed in Django 5.2).
            if isinstance(self.base_field, DecimalField):
                return [self.base_field.get_db_prep_save(i, connection) for i in value]
            return [self.base_field.get_db_prep_value(i, connection, prepared=False) for i in value]
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path == "django_mongodb_backend.fields.array.ArrayField":
            path = "django_mongodb_backend.fields.ArrayField"
        kwargs["base_field"] = self.base_field.clone()
        if self.max_size is not None:
            kwargs["max_size"] = self.max_size
        if self.size is not None:
            kwargs["size"] = self.size
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, str):
            # Assume value is being deserialized.
            vals = json.loads(value)
            value = [self.base_field.to_python(val) for val in vals]
        return value

    def _from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return [self.base_field.from_db_value(item, expression, connection) for item in value]

    def value_to_string(self, obj):
        values = []
        vals = self.value_from_object(obj)
        base_field = self.base_field

        for val in vals:
            if val is None:
                values.append(None)
            else:
                obj = AttributeSetter(base_field.attname, val)
                values.append(base_field.value_to_string(obj))
        return json.dumps(values)

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        if "_" not in name:
            try:
                index = int(name)
            except ValueError:
                pass
            else:
                return IndexTransformFactory(index, self.base_field)
        try:
            start, end = name.split("_")
            start = int(start)
            end = int(end)
        except ValueError:
            pass
        else:
            return SliceTransformFactory(start, end)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        for index, part in enumerate(value):
            try:
                self.base_field.validate(part, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                ) from None
        if isinstance(self.base_field, ArrayField) and len({len(i) for i in value}) > 1:
            raise exceptions.ValidationError(
                self.error_messages["nested_array_mismatch"],
                code="nested_array_mismatch",
            )

    def run_validators(self, value):
        super().run_validators(value)
        for index, part in enumerate(value):
            try:
                self.base_field.run_validators(part)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                ) from None

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": SimpleArrayField,
                "base_field": self.base_field.formfield(),
                "max_length": self.max_size,
                "length": self.size,
                **kwargs,
            }
        )


class Array(Func):
    def as_mql(self, compiler, connection):
        return [expr.as_mql(compiler, connection) for expr in self.get_source_expressions()]


class ArrayRHSMixin:
    def __init__(self, lhs, rhs):
        if isinstance(rhs, tuple | list):
            expressions = []
            for value in rhs:
                if not hasattr(value, "resolve_expression"):
                    field = lhs.output_field
                    value = Value(field.base_field.get_prep_value(value))
                expressions.append(value)
            rhs = Array(*expressions)
        super().__init__(lhs, rhs)


@ArrayField.register_lookup
class ArrayContains(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "contains"

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {
            "$and": [
                {"$ne": [lhs_mql, None]},
                {"$ne": [value, None]},
                {"$setIsSubset": [value, lhs_mql]},
            ]
        }


@ArrayField.register_lookup
class ArrayContainedBy(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "contained_by"

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {
            "$and": [
                {"$ne": [lhs_mql, None]},
                {"$ne": [value, None]},
                {"$setIsSubset": [lhs_mql, value]},
            ]
        }


@ArrayField.register_lookup
class ArrayExact(ArrayRHSMixin, Exact):
    pass


@ArrayField.register_lookup
class ArrayOverlap(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "overlap"

    def get_subquery_wrapping_pipeline(self, compiler, connection, field_name, expr):
        return [
            {
                "$facet": {
                    "group": [
                        {"$project": {"tmp_name": expr.as_mql(compiler, connection)}},
                        {
                            "$unwind": "$tmp_name",
                        },
                        {
                            "$group": {
                                "_id": None,
                                "tmp_name": {"$addToSet": "$tmp_name"},
                            }
                        },
                    ]
                }
            },
            {
                "$project": {
                    field_name: {
                        "$ifNull": [
                            {
                                "$getField": {
                                    "input": {"$arrayElemAt": ["$group", 0]},
                                    "field": "tmp_name",
                                }
                            },
                            [],
                        ]
                    }
                }
            },
        ]

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {
            "$and": [{"$ne": [lhs_mql, None]}, {"$size": {"$setIntersection": [value, lhs_mql]}}]
        }


@ArrayField.register_lookup
class ArrayLenTransform(Transform):
    lookup_name = "len"
    output_field = IntegerField()

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return {"$cond": {"if": {"$eq": [lhs_mql, None]}, "then": None, "else": {"$size": lhs_mql}}}


@ArrayField.register_lookup
class ArrayInLookup(In):
    def get_prep_lookup(self):
        values = super().get_prep_lookup()
        if hasattr(values, "resolve_expression"):
            return values
        # process_rhs() expects hashable values, so convert lists to tuples.
        prepared_values = []
        for value in values:
            if hasattr(value, "resolve_expression"):
                prepared_values.append(value)
            else:
                prepared_values.append(tuple(value))
        return prepared_values


class IndexTransform(Transform):
    def __init__(self, index, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index
        self.base_field = base_field

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return {"$arrayElemAt": [lhs_mql, self.index]}

    @property
    def output_field(self):
        return self.base_field


class IndexTransformFactory:
    def __init__(self, index, base_field):
        self.index = index
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return IndexTransform(self.index, self.base_field, *args, **kwargs)


class SliceTransform(Transform):
    def __init__(self, start, end, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start = start
        self.end = end

    def as_mql(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return {"$slice": [lhs_mql, self.start, self.end]}


class SliceTransformFactory:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, *args, **kwargs):
        return SliceTransform(self.start, self.end, *args, **kwargs)



from django.core import checks
from django.db.backends.base.validation import BaseDatabaseValidation


class DatabaseValidation(BaseDatabaseValidation):
    prohibited_fields = {"AutoField", "BigAutoField", "SmallAutoField"}

    def check_field_type(self, field, field_type):
        """Prohibit AutoField on MongoDB."""
        errors = []
        if field.get_internal_type() in self.prohibited_fields:
            errors.append(
                checks.Error(
                    f"{self.connection.display_name} does not support {field.__class__.__name__}.",
                    obj=field,
                    hint="Use django_mongodb_backend.fields.ObjectIdAutoField instead.",
                    id="mongodb.E001",
                )
            )
        return errors



from django import forms

from .models import Author, Book


class AuthorForm(forms.ModelForm):
    class Meta:
        fields = "__all__"
        model = Author


class BookForm(forms.ModelForm):
    class Meta:
        fields = "__all__"
        model = Book



from bson import ObjectId
from bson.errors import InvalidId
from django.core.exceptions import ValidationError
from django.forms import Field
from django.utils.translation import gettext_lazy as _


class ObjectIdField(Field):
    default_error_messages = {
        "invalid": _("Enter a valid Object Id."),
    }

    def prepare_value(self, value):
        if isinstance(value, ObjectId):
            return str(value)
        return value

    def to_python(self, value):
        value = super().to_python(value)
        if value in self.empty_values:
            return None
        if not isinstance(value, ObjectId):
            try:
                value = ObjectId(value)
            except InvalidId:
                raise ValidationError(self.error_messages["invalid"], code="invalid") from None
        return value



import operator
from datetime import timedelta

from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import models
from django.db.models import (
    Exists,
    ExpressionWrapper,
    F,
    Max,
    OuterRef,
    Sum,
)
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import EmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel

from .models import (
    Address,
    Author,
    Book,
    Data,
    Holder,
    Library,
    NestedData,
)
from .utils import truncate_ms


class MethodTests(SimpleTestCase):
    def test_deconstruct(self):
        field = EmbeddedModelField("Data", null=True)
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "django_mongodb_backend.fields.EmbeddedModelField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"embedded_model": "Data", "null": True})

    def test_get_db_prep_save_invalid(self):
        msg = "Expected instance of type <class 'model_fields_.models.Data'>, " "not <class 'int'>."
        with self.assertRaisesMessage(TypeError, msg):
            Holder(data=42).save()

    def test_validate(self):
        obj = Holder(data=Data(integer=None))
        # This isn't quite right because "integer" is the subfield of data
        # that's non-null.
        msg = "{'data': ['This field cannot be null.']}"
        with self.assertRaisesMessage(ValidationError, msg):
            obj.full_clean()


class ModelTests(TestCase):
    def test_save_load(self):
        Holder.objects.create(data=Data(integer="5"))
        obj = Holder.objects.get()
        self.assertIsInstance(obj.data, Data)
        # get_prep_value() is called, transforming string to int.
        self.assertEqual(obj.data.integer, 5)
        # Primary keys should not be populated...
        self.assertEqual(obj.data.id, None)
        # ... unless set explicitly.
        obj.data.id = obj.id
        obj.save()
        obj = Holder.objects.get()
        self.assertEqual(obj.data.id, obj.id)

    def test_save_load_null(self):
        Holder.objects.create(data=None)
        obj = Holder.objects.get()
        self.assertIsNone(obj.data)

    def test_save_load_json(self):
        obj = Holder.objects.create(data=Data(json_value={"a": 1}))
        obj.refresh_from_db()
        self.assertEqual(obj.data.json_value, {"a": 1})

    def test_pre_save(self):
        """Field.pre_save() is called on embedded model fields."""
        obj = Holder.objects.create(data=Data())
        auto_now = truncate_ms(obj.data.auto_now)
        auto_now_add = truncate_ms(obj.data.auto_now_add)
        # auto_now and auto_now_add may differ by a millisecond since they
        # aren't generated simultaneously.
        self.assertAlmostEqual(auto_now, auto_now_add, delta=timedelta(microseconds=1000))
        # save() updates auto_now but not auto_now_add.
        obj.save()
        self.assertEqual(truncate_ms(obj.data.auto_now_add), auto_now_add)
        auto_now_two = obj.data.auto_now
        self.assertGreater(auto_now_two, obj.data.auto_now_add)
        # And again, save() updates auto_now but not auto_now_add.
        obj = Holder.objects.get()
        obj.save()
        self.assertEqual(obj.data.auto_now_add, auto_now_add)
        self.assertGreater(obj.data.auto_now, auto_now_two)


class QueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.objs = [
            Holder.objects.create(
                data=Data(
                    integer=x,
                    decimal=f"{x}.5",
                    nested_data=NestedData(decimal=f"{x}.5"),
                )
            )
            for x in range(6)
        ]

    def test_exact(self):
        self.assertCountEqual(Holder.objects.filter(data__integer=3), [self.objs[3]])

    def test_lt(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__lt=3), self.objs[:3])

    def test_lte(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__lte=3), self.objs[:4])

    def test_gt(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__gt=3), self.objs[4:])

    def test_gte(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__gte=3), self.objs[3:])

    def test_range(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__range=(2, 4)), self.objs[2:5])

    def test_exact_decimal(self):
        # EmbeddedModelField lookups call
        # DatabaseOperations.adapt__<foo>field_value().
        self.assertCountEqual(Holder.objects.filter(data__decimal="3.5"), [self.objs[3]])

    def test_lt_decimal(self):
        self.assertCountEqual(Holder.objects.filter(data__decimal__lt="3"), self.objs[0:3])

    def test_exact_decimal_nested(self):
        self.assertCountEqual(
            Holder.objects.filter(data__nested_data__decimal="3.5"), [self.objs[3]]
        )

    def test_order_by_embedded_field(self):
        qs = Holder.objects.filter(data__integer__gt=3).order_by("-data__integer")
        self.assertSequenceEqual(qs, list(reversed(self.objs[4:])))

    def test_embedded_json_field_lookups(self):
        objs = [
            Holder.objects.create(
                data=Data(json_value={"field1": i * 5, "field2": {"0": {"value": list(range(i))}}})
            )
            for i in range(4)
        ]
        self.assertCountEqual(
            Holder.objects.filter(data__json_value__field2__0__value__0=0),
            objs[1:],
        )
        self.assertCountEqual(
            Holder.objects.filter(data__json_value__field2__0__value__1=1),
            objs[2:],
        )
        self.assertCountEqual(Holder.objects.filter(data__json_value__field2__0__value__1=5), [])
        self.assertCountEqual(Holder.objects.filter(data__json_value__field1__lt=100), objs)
        self.assertCountEqual(Holder.objects.filter(data__json_value__field1__gt=100), [])
        self.assertCountEqual(
            Holder.objects.filter(
                data__json_value__field1__gte=5, data__json_value__field1__lte=10
            ),
            objs[1:3],
        )

    def test_order_and_group_by_embedded_field(self):
        # Create and sort test data by `data__integer`.
        expected_objs = sorted(
            (Holder.objects.create(data=Data(integer=x)) for x in range(6)),
            key=lambda x: x.data.integer,
        )
        # Group by `data__integer + 5` and get the latest `data__auto_now`
        # datetime.
        qs = (
            Holder.objects.annotate(
                group=ExpressionWrapper(F("data__integer") + 5, output_field=models.IntegerField()),
            )
            .values("group")
            .annotate(max_auto_now=Max("data__auto_now"))
            .order_by("data__integer")
        )
        # Each unique `data__integer` is correctly grouped and annotated.
        self.assertSequenceEqual(
            [{**e, "max_auto_now": e["max_auto_now"]} for e in qs],
            [
                {"group": e.data.integer + 5, "max_auto_now": truncate_ms(e.data.auto_now)}
                for e in expected_objs
            ],
        )

    def test_order_and_group_by_embedded_field_annotation(self):
        # Create repeated `data__integer` values.
        [Holder.objects.create(data=Data(integer=x)) for x in range(6)]
        # Group by `data__integer` and compute the sum of occurrences.
        qs = (
            Holder.objects.values("data__integer")
            .annotate(sum=Sum("data__integer"))
            .order_by("sum")
        )
        # The sum is twice the integer values since each appears twice.
        self.assertQuerySetEqual(qs, [0, 2, 4, 6, 8, 10], operator.itemgetter("sum"))

    def test_nested(self):
        obj = Book.objects.create(
            author=Author(name="Shakespeare", age=55, address=Address(city="NYC", state="NY"))
        )
        self.assertCountEqual(Book.objects.filter(author__address__city="NYC"), [obj])


class ArrayFieldTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.book = Book.objects.create(
            author=Author(
                name="Shakespeare",
                age=55,
                skills=["writing", "editing"],
                address=Address(city="NYC", state="NY", tags=["home", "shipping"]),
            ),
        )

    def test_contains(self):
        self.assertCountEqual(Book.objects.filter(author__skills__contains=["nonexistent"]), [])
        self.assertCountEqual(
            Book.objects.filter(author__skills__contains=["writing"]), [self.book]
        )
        # Nested
        self.assertCountEqual(
            Book.objects.filter(author__address__tags__contains=["nonexistent"]), []
        )
        self.assertCountEqual(
            Book.objects.filter(author__address__tags__contains=["home"]), [self.book]
        )

    def test_contained_by(self):
        self.assertCountEqual(
            Book.objects.filter(author__skills__contained_by=["writing", "publishing"]), []
        )
        self.assertCountEqual(
            Book.objects.filter(author__skills__contained_by=["writing", "editing", "publishing"]),
            [self.book],
        )
        # Nested
        self.assertCountEqual(
            Book.objects.filter(author__address__tags__contained_by=["home", "work"]), []
        )
        self.assertCountEqual(
            Book.objects.filter(author__address__tags__contained_by=["home", "work", "shipping"]),
            [self.book],
        )

    def test_len(self):
        self.assertCountEqual(Book.objects.filter(author__skills__len=1), [])
        self.assertCountEqual(Book.objects.filter(author__skills__len=2), [self.book])
        # Nested
        self.assertCountEqual(Book.objects.filter(author__address__tags__len=1), [])
        self.assertCountEqual(Book.objects.filter(author__address__tags__len=2), [self.book])


class InvalidLookupTests(SimpleTestCase):
    def test_invalid_field(self):
        msg = "Author has no field named 'first_name'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Book.objects.filter(author__first_name="Bob")

    def test_invalid_field_nested(self):
        msg = "Address has no field named 'floor'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Book.objects.filter(author__address__floor="NYC")

    def test_invalid_lookup(self):
        msg = "Unsupported lookup 'foo' for CharField 'city'."
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Book.objects.filter(author__address__city__foo="NYC")

    def test_invalid_lookup_with_suggestions(self):
        msg = (
            "Unsupported lookup '{lookup}' for CharField 'name', "
            "perhaps you meant {suggested_lookups}?"
        )
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="exactly", suggested_lookups="exact or iexact")
        ):
            Book.objects.filter(author__name__exactly="NYC")
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="gti", suggested_lookups="gt or gte")
        ):
            Book.objects.filter(author__name__gti="NYC")
        with self.assertRaisesMessage(
            FieldDoesNotExist, msg.format(lookup="is_null", suggested_lookups="isnull")
        ):
            Book.objects.filter(author__name__is_null="NYC")


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_no_relational_fields(self):
        class Target(EmbeddedModel):
            key = models.ForeignKey("MyModel", models.CASCADE)

        class MyModel(models.Model):
            field = EmbeddedModelField(Target)

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "django_mongodb_backend.embedded_model.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg, "Embedded models cannot have relational fields (Target.key is a ForeignKey)."
        )

    def test_embedded_model_subclass(self):
        class Target(models.Model):
            pass

        class MyModel(models.Model):
            field = EmbeddedModelField(Target)

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "django_mongodb_backend.embedded_model.E002")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Embedded models must be a subclass of django_mongodb_backend.models.EmbeddedModel.",
        )


class SubqueryExistsTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        address1 = Address(city="New York", state="NY", zip_code=10001)
        address2 = Address(city="Boston", state="MA", zip_code=20002)
        author1 = Author(name="Alice", age=30, address=address1)
        author2 = Author(name="Bob", age=40, address=address2)
        book1 = Book.objects.create(name="Book 1", author=author1)
        book2 = Book.objects.create(name="Book 2", author=author2)
        Book.objects.create(name="Book 3", author=author2)
        Book.objects.create(name="Book 4", author=author2)
        Book.objects.create(name="Book 5", author=author1)
        library1 = Library.objects.create(name="Library 1", best_seller="Book 1")
        library2 = Library.objects.create(name="Library 2", best_seller="Book 1")
        library1.books.add(book1, book2)
        library2.books.add(book2)

    def test_exists_subquery(self):
        subquery = Book.objects.filter(
            author__name=OuterRef("author__name"), author__address__city="Boston"
        )
        qs = Book.objects.filter(Exists(subquery)).order_by("name")
        self.assertQuerySetEqual(qs, ["Book 2", "Book 3", "Book 4"], lambda book: book.name)

    def test_in_subquery(self):
        names = Book.objects.filter(author__age__gt=35).values("author__name")
        qs = Book.objects.filter(author__name__in=names).order_by("name")
        self.assertQuerySetEqual(qs, ["Book 2", "Book 3", "Book 4"], lambda book: book.name)

    def test_exists_with_foreign_object(self):
        subquery = Library.objects.filter(best_seller=OuterRef("name"))
        qs = Book.objects.filter(Exists(subquery))
        self.assertEqual(qs.first().name, "Book 1")

    def test_foreign_field_with_range(self):
        qs = Library.objects.filter(books__author__age__range=(25, 35))
        self.assertEqual(qs.first().name, "Library 1")



from .fields import (
    EmbeddedModelField,
    ObjectIdField,
    SimpleArrayField,
    SplitArrayField,
    SplitArrayWidget,
)

__all__ = [
    "EmbeddedModelField",
    "SimpleArrayField",
    "SplitArrayField",
    "SplitArrayWidget",
    "ObjectIdField",
]



import signal
from unittest import mock

from django.db import connection
from django.test import SimpleTestCase

from django_mongodb_backend.client import DatabaseClient


class MongoDbshellTests(SimpleTestCase):
    def settings_to_cmd_args_env(self, settings_dict, parameters=None):
        if parameters is None:
            parameters = []
        return DatabaseClient.settings_to_cmd_args_env(settings_dict, parameters)

    def test_fails_with_keyerror_on_incomplete_config(self):
        with self.assertRaises(KeyError):
            self.settings_to_cmd_args_env({})

    def test_basic_params_specified_in_settings(self):
        for options_parameters in [(None, None, None), ("value1", "value2", True)]:
            with self.subTest(keys=options_parameters):
                authentication_database, authentication_mechanism, retry_writes = options_parameters
                if authentication_database is not None:
                    expected_args = [
                        "mongosh",
                        "--host",
                        "somehost",
                        "--port",
                        444,
                        "--username",
                        "someuser",
                        "--password",
                        "somepassword",
                        "--retryWrites",
                        "true",
                        "somedbname",
                    ]
                else:
                    expected_args = [
                        "mongosh",
                        "--host",
                        "somehost",
                        "--port",
                        444,
                        "--username",
                        "someuser",
                        "--password",
                        "somepassword",
                        "somedbname",
                    ]

                self.assertEqual(
                    self.settings_to_cmd_args_env(
                        {
                            "NAME": "somedbname",
                            "USER": "someuser",
                            "PASSWORD": "somepassword",
                            "HOST": "somehost",
                            "PORT": 444,
                            "OPTIONS": {
                                "authenticationDatabase": authentication_database,
                                "authenticationMechanism": authentication_mechanism,
                                "retryWrites": retry_writes,
                            },
                        }
                    ),
                    (expected_args, None),
                )

    def test_options_override_settings_proper_values(self):
        settings_port = 444
        options_port = 555
        self.assertNotEqual(settings_port, options_port, "test pre-req")
        expected_args = [
            "mongosh",
            "--host",
            "settinghost",
            "--port",
            444,
            "--username",
            "settinguser",
            "--password",
            "settingpassword",
            "settingdbname",
        ]
        expected_env = None

        self.assertEqual(
            self.settings_to_cmd_args_env(
                {
                    "NAME": "settingdbname",
                    "USER": "settinguser",
                    "PASSWORD": "settingpassword",
                    "HOST": "settinghost",
                    "PORT": settings_port,
                    "OPTIONS": {"port": options_port},
                }
            ),
            (expected_args, expected_env),
        )

    def test_sigint_handler(self):
        """SIGINT is ignored in Python and passed to Mongodb to abort queries."""

        def _mock_subprocess_run(*args, **kwargs):  # noqa: ARG001
            handler = signal.getsignal(signal.SIGINT)
            self.assertEqual(handler, signal.SIG_IGN)

        sigint_handler = signal.getsignal(signal.SIGINT)
        # The default handler isn't SIG_IGN.
        self.assertNotEqual(sigint_handler, signal.SIG_IGN)
        with mock.patch("subprocess.run", new=_mock_subprocess_run):
            connection.client.runshell([])
        # dbshell restores the original handler.
        self.assertEqual(sigint_handler, signal.getsignal(signal.SIGINT))



from django.db import models

from django_mongodb_backend.fields import EmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel


class Address(EmbeddedModel):
    po_box = models.CharField(max_length=50, blank=True, verbose_name="PO Box")
    city = models.CharField(max_length=20)
    state = models.CharField(max_length=2)
    zip_code = models.IntegerField()


class Author(models.Model):
    name = models.CharField(max_length=10)
    age = models.IntegerField()
    address = EmbeddedModelField(Address)
    billing_address = EmbeddedModelField(Address, blank=True, null=True)


class Publisher(EmbeddedModel):
    name = models.CharField(max_length=50)
    address = EmbeddedModelField(Address)


class Book(models.Model):
    title = models.CharField(max_length=50)
    publisher = EmbeddedModelField(Publisher)



from django.db import models

from django_mongodb_backend.fields import ObjectIdAutoField, ObjectIdField


class Author(models.Model):
    name = models.CharField(max_length=10)

    def __str__(self):
        return self.name


class Book(models.Model):
    title = models.CharField(max_length=10)
    author = models.ForeignKey(Author, models.CASCADE)

    def __str__(self):
        return self.title


class Tag(models.Model):
    name = models.CharField(max_length=10)
    parent = models.ForeignKey(
        "self",
        models.SET_NULL,
        blank=True,
        null=True,
        related_name="children",
    )
    group_id = ObjectIdField(null=True)

    def __str__(self):
        return self.name


class Order(models.Model):
    id = ObjectIdAutoField(primary_key=True)
    name = models.CharField(max_length=12, null=True, default="")

    class Meta:
        ordering = ("pk",)

    def __str__(self):
        return str(self.pk)


class OrderItem(models.Model):
    order = models.ForeignKey(Order, models.CASCADE, related_name="items")
    status = ObjectIdField(null=True)

    class Meta:
        ordering = ("pk",)

    def __str__(self):
        return str(self.pk)



from django.test import TestCase

from .models import Author, Book


class MQLTests(TestCase):
    def test_all(self):
        with self.assertNumQueries(1) as ctx:
            list(Author.objects.all())
        query = ctx.captured_queries[0]["sql"]
        self.assertEqual(query, "db.queries__author.aggregate([{'$match': {'$expr': {}}}])")

    def test_join(self):
        with self.assertNumQueries(1) as ctx:
            list(Book.objects.filter(author__name="Bob"))
        query = ctx.captured_queries[0]["sql"]
        self.assertEqual(
            query,
            "db.queries__book.aggregate(["
            "{'$lookup': {'from': 'queries__author', "
            "'let': {'parent__field__0': '$author_id'}, "
            "'pipeline': [{'$match': {'$expr': "
            "{'$and': [{'$eq': ['$$parent__field__0', '$_id']}]}}}], 'as': 'queries__author'}}, "
            "{'$unwind': '$queries__author'}, "
            "{'$match': {'$expr': {'$eq': ['$queries__author.name', 'Bob']}}}])",
        )



from django.apps import apps

from django_mongodb_backend.models import EmbeddedModel


class MongoRouter:
    def allow_migrate(self, db, app_label, model_name=None, **hints):
        """
        EmbeddedModels don't have their own collection and must be ignored by
        dumpdata.
        """
        if not model_name:
            return None
        try:
            model = apps.get_model(app_label, model_name)
        except LookupError:
            return None
        return False if issubclass(model, EmbeddedModel) else None



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from __future__ import annotations

import sys
from importlib.metadata import version as _version
from pathlib import Path

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(str((Path(__file__).parent / "_ext").resolve()))

project = "Django MongoDB Backend"
copyright = "2025, The MongoDB Python Team"
author = "The MongoDB Python Team"
release = _version("django_mongodb_backend")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

extensions = [
    "djangodocs",
    "sphinx.ext.intersphinx",
]

# templates_path = ["_templates"]
exclude_patterns = []

intersphinx_mapping = {
    "django": (
        "https://docs.djangoproject.com/en/5.1/",
        "https://docs.djangoproject.com/en/5.1/_objects/",
    ),
    "mongodb": ("https://www.mongodb.com/docs/languages/python/django-mongodb/v5.1/", None),
    "pymongo": ("https://pymongo.readthedocs.io/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
}

root_doc = "contents"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



from django.test import SimpleTestCase

from django_mongodb_backend.routers import MongoRouter


class TestRouter(SimpleTestCase):
    def setUp(self):
        self.router = MongoRouter()

    def test_no_model(self):
        self.assertIsNone(self.router.allow_migrate("db", "models_"))

    def test_regular_model(self):
        self.assertIsNone(self.router.allow_migrate("db", "models_", "plainmodel"))

    def test_nonexistent_model(self):
        self.assertIsNone(self.router.allow_migrate("db", "models_", "nonexistentmodel"))

    def test_embedded_model(self):
        self.assertIs(self.router.allow_migrate("db", "models_", "embed"), False)



from itertools import chain

from django.core.exceptions import FieldDoesNotExist
from django.db import connections
from django.db.models import QuerySet
from django.db.models.query import RawModelIterable as BaseRawModelIterable
from django.db.models.query import RawQuerySet as BaseRawQuerySet
from django.db.models.sql.query import RawQuery as BaseRawQuery


class MongoQuerySet(QuerySet):
    def raw_aggregate(self, pipeline, using=None):
        return RawQuerySet(pipeline, model=self.model, using=using)


class RawQuerySet(BaseRawQuerySet):
    def __init__(self, pipeline, model=None, using=None):
        super().__init__(pipeline, model=model, using=using)
        self.query = RawQuery(pipeline, using=self.db, model=self.model)
        # Override the superclass's columns property which relies on PEP 249's
        # cursor.description. Instead, RawModelIterable will set the columns
        # based on the keys in the first result.
        self.columns = None

    def iterator(self):
        yield from RawModelIterable(self)


class RawQuery(BaseRawQuery):
    def __init__(self, pipeline, using, model):
        self.pipeline = pipeline
        super().__init__(sql=None, using=using)
        self.model = model

    def _execute_query(self):
        connection = connections[self.using]
        collection = connection.get_collection(self.model._meta.db_table)
        self.cursor = collection.aggregate(self.pipeline)

    def __str__(self):
        return str(self.pipeline)


class RawModelIterable(BaseRawModelIterable):
    def __iter__(self):
        """
        This is copied from the superclass except for the part that sets
        self.queryset.columns from the first result.
        """
        db = self.queryset.db
        query = self.queryset.query
        connection = connections[db]
        compiler = connection.ops.compiler("SQLCompiler")(query, connection, db)
        query_iterator = iter(query)
        try:
            # Get the columns from the first result.
            try:
                first_result = next(query_iterator)
            except StopIteration:
                # No results.
                return
            self.queryset.columns = list(first_result.keys())
            # Reset the iterator to include the first item.
            query_iterator = self._make_result(chain([first_result], query_iterator))
            (
                model_init_names,
                model_init_pos,
                annotation_fields,
            ) = self.queryset.resolve_model_init_order()
            model_cls = self.queryset.model
            if model_cls._meta.pk.attname not in model_init_names:
                raise FieldDoesNotExist("Raw query must include the primary key")
            fields = [self.queryset.model_fields.get(c) for c in self.queryset.columns]
            converters = compiler.get_converters(
                [f.get_col(f.model._meta.db_table) if f else None for f in fields]
            )
            if converters:
                query_iterator = compiler.apply_converters(query_iterator, converters)
            for values in query_iterator:
                # Associate fields to values
                model_init_values = [values[pos] for pos in model_init_pos]
                instance = model_cls.from_db(db, model_init_names, model_init_values)
                if annotation_fields:
                    for column, pos in annotation_fields:
                        setattr(instance, column, values[pos])
                yield instance
        finally:
            query.cursor.close()

    def _make_result(self, query):
        """
        Convert documents (dictionaries) to tuples as expected by the rest
        of __iter__().
        """
        for result in query:
            yield tuple(result.get(key) for key in self.queryset.columns)



from django.db import NotSupportedError
from django.db.models.manager import BaseManager

from .queryset import MongoQuerySet


class MongoManager(BaseManager.from_queryset(MongoQuerySet)):
    pass


class EmbeddedModelManager(BaseManager):
    """
    Prevent all queryset operations on embedded models since they don't have
    their own collection.

    Raise a helpful error message for some basic QuerySet methods. Subclassing
    BaseManager means that other methods raise, e.g. AttributeError:
    'EmbeddedModelManager' object has no attribute 'update_or_create'".
    """

    def all(self):
        raise NotSupportedError("EmbeddedModels cannot be queried.")

    def get(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be queried.")

    def filter(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be queried.")

    def create(self, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be created.")

    def update(self, *args, **kwargs):
        raise NotSupportedError("EmbeddedModels cannot be updated.")

    def delete(self):
        raise NotSupportedError("EmbeddedModels cannot be deleted.")



from django.db import migrations, models

import django_mongodb_backend


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CharTextArrayIndexModel",
            fields=[
                (
                    "id",
                    django_mongodb_backend.fields.ObjectIdAutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "char",
                    django_mongodb_backend.fields.ArrayField(
                        models.CharField(max_length=10), db_index=True, max_size=100
                    ),
                ),
                ("char2", models.CharField(max_length=11, db_index=True)),
                (
                    "text",
                    django_mongodb_backend.fields.ArrayField(models.TextField(), db_index=True),
                ),
            ],
            options={},
            bases=(models.Model,),
        ),
    ]



import operator

from django.db import NotSupportedError, connection
from django.db.models import Index, Q
from django.test import TestCase

from .models import Article


class PartialIndexTests(TestCase):
    def assertAddRemoveIndex(self, editor, model, index):
        editor.add_index(index=index, model=model)
        self.assertIn(
            index.name,
            connection.introspection.get_constraints(
                cursor=None,
                table_name=model._meta.db_table,
            ),
        )
        editor.remove_index(index=index, model=model)

    def test_not_supported(self):
        msg = "MongoDB does not support the 'isnull' lookup in indexes."
        with connection.schema_editor() as editor, self.assertRaisesMessage(NotSupportedError, msg):
            Index(
                name="test",
                fields=["headline"],
                condition=Q(pk__isnull=True),
            )._get_condition_mql(Article, schema_editor=editor)

    def test_negated_not_supported(self):
        msg = "MongoDB does not support the '~' operator in indexes."
        with self.assertRaisesMessage(NotSupportedError, msg), connection.schema_editor() as editor:
            Index(
                name="test",
                fields=["headline"],
                condition=~Q(pk__isnull=True),
            )._get_condition_mql(Article, schema_editor=editor)

    def test_xor_not_supported(self):
        msg = "MongoDB does not support the '^' operator lookup in indexes."
        with self.assertRaisesMessage(NotSupportedError, msg), connection.schema_editor() as editor:
            Index(
                name="test",
                fields=["headline"],
                condition=Q(pk__isnull=True) ^ Q(pk__isnull=False),
            )._get_condition_mql(Article, schema_editor=editor)

    def test_operations(self):
        operators = (
            ("gt", "$gt"),
            ("gte", "$gte"),
            ("lt", "$lt"),
            ("lte", "$lte"),
        )
        for op, mongo_operator in operators:
            with self.subTest(operator=op), connection.schema_editor() as editor:
                index = Index(
                    name="test",
                    fields=["headline"],
                    condition=Q(**{f"number__{op}": 3}),
                )
                self.assertEqual(
                    {"number": {mongo_operator: 3}},
                    index._get_condition_mql(Article, schema_editor=editor),
                )
                self.assertAddRemoveIndex(editor, Article, index)

    def test_composite_index(self):
        with connection.schema_editor() as editor:
            index = Index(
                name="test",
                fields=["headline"],
                condition=Q(number__gte=3) & (Q(body__gt="test1") | Q(body__in=["A", "B"])),
            )
            self.assertEqual(
                index._get_condition_mql(Article, schema_editor=editor),
                {
                    "$and": [
                        {"number": {"$gte": 3}},
                        {"$or": [{"body": {"$gt": "test1"}}, {"body": {"$in": ["A", "B"]}}]},
                    ]
                },
            )
            self.assertAddRemoveIndex(editor, Article, index)

    def test_composite_op_index(self):
        operators = (
            (operator.or_, "$or"),
            (operator.and_, "$and"),
        )
        for op, mongo_operator in operators:
            with self.subTest(operator=op), connection.schema_editor() as editor:
                index = Index(
                    name="test",
                    fields=["headline"],
                    condition=op(Q(number__gte=3), Q(body__gt="test1")),
                )
                self.assertEqual(
                    {mongo_operator: [{"number": {"$gte": 3}}, {"body": {"$gt": "test1"}}]},
                    index._get_condition_mql(Article, schema_editor=editor),
                )
                self.assertAddRemoveIndex(editor, Article, index)



from django.conf import settings
from django.core.cache import caches
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, connections, router

from django_mongodb_backend.cache import MongoDBCache


class Command(BaseCommand):
    help = "Creates the collections needed to use the MongoDB cache backend."
    requires_system_checks = []

    def add_arguments(self, parser):
        parser.add_argument(
            "args",
            metavar="collection_name",
            nargs="*",
            help="Optional collections names. Otherwise, settings.CACHES is "
            "used to find cache collections.",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help="Nominates a database onto which the cache collections will be "
            'installed. Defaults to the "default" database.',
        )

    def handle(self, *collection_names, **options):
        db = options["database"]
        self.verbosity = options["verbosity"]
        if collection_names:
            # Legacy behavior, collection_name specified as argument
            for collection_name in collection_names:
                self.check_collection(db, collection_name)
        else:
            for cache_alias in settings.CACHES:
                cache = caches[cache_alias]
                if isinstance(cache, MongoDBCache):
                    self.check_collection(db, cache._collection_name)

    def check_collection(self, database, collection_name):
        cache = MongoDBCache(collection_name, {})
        if not router.allow_migrate_model(database, cache.cache_model_class):
            return
        connection = connections[database]
        if cache._collection_name in connection.introspection.table_names():
            if self.verbosity > 0:
                self.stdout.write("Cache collection '%s' already exists." % cache._collection_name)
            return
        cache.create_indexes()



from django.db.backends.base.features import BaseDatabaseFeatures
from django.utils.functional import cached_property


class DatabaseFeatures(BaseDatabaseFeatures):
    minimum_database_version = (6, 0)
    allow_sliced_subqueries_with_in = False
    allows_multiple_constraints_on_same_fields = False
    can_create_inline_fk = False
    can_introspect_foreign_keys = False
    can_return_rows_from_bulk_insert = True
    greatest_least_ignores_nulls = True
    has_json_object_function = False
    has_native_json_field = True
    supports_boolean_expr_in_select_clause = True
    supports_collation_on_charfield = False
    supports_column_check_constraints = False
    supports_date_lookup_using_string = False
    supports_deferrable_unique_constraints = False
    supports_explaining_query_execution = True
    supports_expression_defaults = False
    supports_expression_indexes = False
    supports_foreign_keys = False
    supports_ignore_conflicts = False
    supports_json_field_contains = False
    # BSON Date type doesn't support microsecond precision.
    supports_microsecond_precision = False
    supports_paramstyle_pyformat = False
    supports_select_difference = False
    supports_select_intersection = False
    supports_sequence_reset = False
    supports_slicing_ordering_in_compound = True
    supports_table_check_constraints = False
    supports_temporal_subtraction = True
    # MongoDB stores datetimes in UTC.
    supports_timezones = False
    # Not implemented: https://github.com/mongodb/django-mongodb-backend/issues/7
    supports_transactions = False
    supports_unspecified_pk = True
    uses_savepoints = False

    _django_test_expected_failures = {
        # $concat only supports strings, not int
        "db_functions.text.test_concat.ConcatTests.test_concat_non_str",
        # QuerySet.order_by() with annotation transform doesn't work:
        # "Expression $mod takes exactly 2 arguments. 1 were passed in"
        # https://github.com/django/django/commit/b0ad41198b3e333f57351e3fce5a1fb47f23f376
        "aggregation.tests.AggregateTestCase.test_order_by_aggregate_transform",
        # 'NulledTransform' object has no attribute 'as_mql'.
        "lookup.tests.LookupTests.test_exact_none_transform",
        # "Save with update_fields did not affect any rows."
        "basic.tests.SelectOnSaveTests.test_select_on_save_lying_update",
        # BaseExpression.convert_value() crashes with Decimal128.
        "aggregation.tests.AggregateTestCase.test_combine_different_types",
        "annotations.tests.NonAggregateAnnotationTestCase.test_combined_f_expression_annotation_with_aggregation",
        # Pattern lookups that use regexMatch don't work on JSONField:
        # Unsupported conversion from array to string in $convert
        "model_fields.test_jsonfield.TestQuerying.test_icontains",
        # MongoDB gives ROUND(365, -1)=360 instead of 370 like other databases.
        "db_functions.math.test_round.RoundTests.test_integer_with_negative_precision",
        # Unexpected alias_refcount in alias_map.
        "queries.tests.Queries1Tests.test_order_by_tables",
        # The $sum aggregation returns 0 instead of None for null.
        "aggregation.test_filter_argument.FilteredAggregateTests.test_plain_annotate",
        "aggregation.tests.AggregateTestCase.test_aggregation_default_passed_another_aggregate",
        "aggregation.tests.AggregateTestCase.test_annotation_expressions",
        "aggregation.tests.AggregateTestCase.test_reverse_fkey_annotate",
        "aggregation_regress.tests.AggregationTests.test_annotation_disjunction",
        "aggregation_regress.tests.AggregationTests.test_decimal_aggregate_annotation_filter",
        # subclasses of BaseDatabaseWrapper may require an is_usable() method
        "backends.tests.BackendTestCase.test_is_usable_after_database_disconnects",
        # Connection creation doesn't follow the usual Django API.
        "backends.tests.ThreadTests.test_pass_connection_between_threads",
        "backends.tests.ThreadTests.test_default_connection_thread_local",
        "test_utils.tests.DisallowedDatabaseQueriesTests.test_disallowed_thread_database_connection",
        # Object of type ObjectId is not JSON serializable.
        "auth_tests.test_views.LoginTest.test_login_session_without_hash_session_key",
        # GenericRelation.value_to_string() assumes integer pk.
        "contenttypes_tests.test_fields.GenericRelationTests.test_value_to_string",
        # icontains doesn't work on ArrayField:
        # Unsupported conversion from array to string in $convert
        "model_fields_.test_arrayfield.QueryingTests.test_icontains",
        # ArrayField's contained_by lookup crashes with Exists: "both operands "
        # of $setIsSubset must be arrays. Second argument is of type: null"
        # https://jira.mongodb.org/browse/SERVER-99186
        "model_fields_.test_arrayfield.QueryingTests.test_contained_by_subquery",
    }
    # $bitAnd, #bitOr, and $bitXor are new in MongoDB 6.3.
    _django_test_expected_failures_bitwise = {
        "expressions.tests.ExpressionOperatorTests.test_lefthand_bitwise_and",
        "expressions.tests.ExpressionOperatorTests.test_lefthand_bitwise_or",
        "expressions.tests.ExpressionOperatorTests.test_lefthand_bitwise_xor",
        "expressions.tests.ExpressionOperatorTests.test_lefthand_bitwise_xor_null",
        "expressions.tests.ExpressionOperatorTests.test_lefthand_bitwise_xor_right_null",
        "expressions.tests.ExpressionOperatorTests.test_lefthand_transformed_field_bitwise_or",
    }

    @cached_property
    def django_test_expected_failures(self):
        expected_failures = super().django_test_expected_failures
        expected_failures.update(self._django_test_expected_failures)
        if not self.is_mongodb_6_3:
            expected_failures.update(self._django_test_expected_failures_bitwise)
        return expected_failures

    django_test_skips = {
        "Database defaults aren't supported by MongoDB.": {
            # bson.errors.InvalidDocument: cannot encode object:
            # <django.db.models.expressions.DatabaseDefault
            "basic.tests.ModelInstanceCreationTests.test_save_primary_with_db_default",
            "constraints.tests.UniqueConstraintTests.test_database_default",
            "field_defaults.tests.DefaultTests",
            "migrations.test_operations.OperationTests.test_add_field_both_defaults",
            "migrations.test_operations.OperationTests.test_add_field_database_default",
            "migrations.test_operations.OperationTests.test_add_field_database_default_special_char_escaping",
            "migrations.test_operations.OperationTests.test_alter_field_add_database_default",
            "migrations.test_operations.OperationTests.test_alter_field_change_blank_nullable_database_default_to_not_null",
            "migrations.test_operations.OperationTests.test_alter_field_change_default_to_database_default",
            "migrations.test_operations.OperationTests.test_alter_field_change_nullable_to_database_default_not_null",
            "migrations.test_operations.OperationTests.test_alter_field_change_nullable_to_decimal_database_default_not_null",
            "schema.tests.SchemaTests.test_db_default_output_field_resolving",
            "schema.tests.SchemaTests.test_rename_keep_db_default",
            "validation.test_unique.PerformUniqueChecksTest.test_unique_db_default",
        },
        "Insert expressions aren't supported.": {
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_now",
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_expressions",
            "expressions.tests.BasicExpressionsTests.test_new_object_create",
            "expressions.tests.BasicExpressionsTests.test_new_object_save",
            "expressions.tests.BasicExpressionsTests.test_object_create_with_aggregate",
            "expressions.tests.BasicExpressionsTests.test_object_create_with_f_expression_in_subquery",
            "expressions.tests.BasicExpressionsTests.test_object_update_unsaved_objects",
            # PI()
            "db_functions.math.test_round.RoundTests.test_decimal_with_precision",
            "db_functions.math.test_round.RoundTests.test_float_with_precision",
        },
        "Check constraints are not supported.": {
            # Constraint.validate() crashes:
            # AttributeError: 'NoneType' object has no attribute 'db_table'
            "constraints.tests.CheckConstraintTests.test_database_default",
            "constraints.tests.CheckConstraintTests.test_validate",
            "constraints.tests.CheckConstraintTests.test_validate_boolean_expressions",
            "constraints.tests.CheckConstraintTests.test_validate_custom_error",
            "constraints.tests.CheckConstraintTests.test_validate_jsonfield_exact",
            "constraints.tests.CheckConstraintTests.test_validate_nullable_field_with_none",
            "constraints.tests.CheckConstraintTests.test_validate_nullable_jsonfield",
            "constraints.tests.CheckConstraintTests.test_validate_pk_field",
            "constraints.tests.CheckConstraintTests.test_validate_rawsql_expressions_noop",
        },
        "MongoDB doesn't rename an index when a field is renamed.": {
            "migrations.test_operations.OperationTests.test_rename_field_index_together",
            "migrations.test_operations.OperationTests.test_rename_field_unique_together",
        },
        "Pattern lookups on UUIDField are not supported.": {
            "model_fields.test_uuid.TestQuerying.test_contains",
            "model_fields.test_uuid.TestQuerying.test_endswith",
            "model_fields.test_uuid.TestQuerying.test_filter_with_expr",
            "model_fields.test_uuid.TestQuerying.test_icontains",
            "model_fields.test_uuid.TestQuerying.test_iendswith",
            "model_fields.test_uuid.TestQuerying.test_iexact",
            "model_fields.test_uuid.TestQuerying.test_istartswith",
            "model_fields.test_uuid.TestQuerying.test_startswith",
        },
        "QuerySet.prefetch_related() is not supported on MongoDB.": {
            "backends.base.test_creation.TestDeserializeDbFromString.test_serialize_db_to_string_base_manager_with_prefetch_related",
            "fixtures.tests.FixtureLoadingTests.test_dumpdata_objects_with_prefetch_related",
            "fixtures.tests.FixtureLoadingTests.test_loading_and_dumping",
            "m2m_through_regress.test_multitable.MultiTableTests.test_m2m_prefetch_proxied",
            "m2m_through_regress.test_multitable.MultiTableTests.test_m2m_prefetch_reverse_proxied",
            "many_to_many.tests.ManyToManyQueryTests.test_prefetch_related_no_queries_optimization_disabled",
            "many_to_many.tests.ManyToManyTests.test_add_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_add_then_remove_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_clear_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_create_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_remove_after_prefetch",
            "many_to_many.tests.ManyToManyTests.test_set_after_prefetch",
            "model_forms.tests.OtherModelFormTests.test_prefetch_related_queryset",
            "prefetch_related.test_prefetch_related_objects.PrefetchRelatedObjectsTests",
            "prefetch_related.test_uuid.UUIDPrefetchRelated",
            "prefetch_related.test_uuid.UUIDPrefetchRelatedLookups",
            "prefetch_related.tests.CustomPrefetchTests",
            "prefetch_related.tests.DefaultManagerTests.test_m2m_then_m2m",
            "prefetch_related.tests.ForeignKeyToFieldTest",
            "prefetch_related.tests.GenericRelationTests.test_traverse_GFK",
            "prefetch_related.tests.LookupOrderingTest.test_order",
            "prefetch_related.tests.MultiDbTests.test_using_is_honored_m2m",
            "prefetch_related.tests.MultiTableInheritanceTest",
            "prefetch_related.tests.PrefetchRelatedTests",
            "prefetch_related.tests.ReadPrefetchedObjectsCacheTests",
            "prefetch_related.tests.Ticket21410Tests",
            "queryset_pickle.tests.PickleabilityTestCase.test_pickle_prefetch_related_with_m2m_and_objects_deletion",
            "serializers.test_json.JsonSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_jsonl.JsonlSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_xml.XmlSerializerTestCase.test_serialize_prefetch_related_m2m",
            "serializers.test_yaml.YamlSerializerTestCase.test_serialize_prefetch_related_m2m",
        },
        "AutoField not supported.": {
            "bulk_create.tests.BulkCreateTests.test_bulk_insert_nullable_fields",
            "introspection.tests.IntrospectionTests.test_sequence_list",
            "lookup.tests.LookupTests.test_filter_by_reverse_related_field_transform",
            "lookup.tests.LookupTests.test_in_ignore_none_with_unhashable_items",
            "m2m_through_regress.tests.ThroughLoadDataTestCase.test_sequence_creation",
            "many_to_many.tests.ManyToManyTests.test_add_remove_invalid_type",
            "many_to_one.tests.ManyToOneTests.test_fk_to_smallautofield",
            "many_to_one.tests.ManyToOneTests.test_fk_to_bigautofield",
            "migrations.test_operations.OperationTests.test_autofield__bigautofield_foreignfield_growth",
            "migrations.test_operations.OperationTests.test_model_with_bigautofield",
            "migrations.test_operations.OperationTests.test_smallfield_autofield_foreignfield_growth",
            "migrations.test_operations.OperationTests.test_smallfield_bigautofield_foreignfield_growth",
            "model_fields.test_autofield.AutoFieldTests",
            "model_fields.test_autofield.BigAutoFieldTests",
            "model_fields.test_autofield.SmallAutoFieldTests",
            "queries.tests.TestInvalidValuesRelation.test_invalid_values",
            "schema.tests.SchemaTests.test_alter_autofield_pk_to_bigautofield_pk",
            "schema.tests.SchemaTests.test_alter_autofield_pk_to_smallautofield_pk",
        },
        "Converters aren't run on returning fields from insert.": {
            # Unsure this is needed for this backend. Can implement by request.
            # https://github.com/django/django/commit/d9de74141e8a920940f1b91ed0a3ccb835b55729
            "custom_pk.tests.CustomPKTests.test_auto_field_subclass_bulk_create",
            "custom_pk.tests.CustomPKTests.test_auto_field_subclass_create",
        },
        "MongoDB does not enforce PositiveIntegerField constraint.": {
            "model_fields.test_integerfield.PositiveIntegerFieldTests.test_negative_values",
        },
        "Test assumes integer primary key.": {
            "custom_managers.tests.CustomManagersRegressTestCase.test_save_clears_annotations_from_base_manager",
            "db_functions.comparison.test_cast.CastTests.test_cast_to_integer_foreign_key",
            "expressions.tests.BasicExpressionsTests.test_nested_subquery_outer_ref_with_autofield",
            "model_fields.test_foreignkey.ForeignKeyTests.test_to_python",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_order_raises_on_non_selected_column",
            "queries.tests.RelatedLookupTypeTests.test_values_queryset_lookup",
            "queries.tests.ValuesSubqueryTests.test_values_in_subquery",
            "sites_tests.tests.CreateDefaultSiteTests.test_no_site_id",
        },
        "Cannot use QuerySet.delete() when querying across multiple collections on MongoDB.": {
            "admin_changelist.tests.ChangeListTests.test_distinct_for_many_to_many_at_second_level_in_search_fields",
            "admin_changelist.tests.ChangeListTests.test_distinct_for_through_m2m_at_second_level_in_list_filter",
            "delete.tests.FastDeleteTests.test_fast_delete_aggregation",
            "delete.tests.FastDeleteTests.test_fast_delete_empty_no_update_can_self_select",
            "delete.tests.FastDeleteTests.test_fast_delete_full_match",
            "delete.tests.FastDeleteTests.test_fast_delete_joined_qs",
            "delete_regress.tests.DeleteTests.test_meta_ordered_delete",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_annotate",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_defer",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_select_related",
            "one_to_one.tests.OneToOneTests.test_o2o_primary_key_delete",
        },
        "Cannot use QuerySet.delete() when a subquery is required.": {
            "custom_managers.tests.CustomManagerTests.test_removal_through_default_m2m_related_manager",
            "custom_managers.tests.CustomManagerTests.test_removal_through_specified_m2m_related_manager",
            "delete_regress.tests.DeleteTests.test_self_reference_with_through_m2m_at_second_level",
            "many_to_many.tests.ManyToManyTests.test_assign",
            "many_to_many.tests.ManyToManyTests.test_assign_ids",
            "many_to_many.tests.ManyToManyTests.test_clear",
            "many_to_many.tests.ManyToManyTests.test_remove",
            "many_to_many.tests.ManyToManyTests.test_reverse_assign_with_queryset",
            "many_to_many.tests.ManyToManyTests.test_set",
            "many_to_many.tests.ManyToManyTests.test_set_existing_different_type",
        },
        "Cannot use QuerySet.update() when querying across multiple collections on MongoDB.": {
            "custom_managers.tests.CustomManagersRegressTestCase.test_refresh_from_db_when_default_manager_filters",
            "expressions.tests.BasicExpressionsTests.test_filter_with_join",
            "model_inheritance.tests.ModelInheritanceDataTests.test_update_works_on_parent_and_child_models_at_once",
            "queries.tests.Queries4Tests.test_ticket7095",
            "queries.tests.Queries5Tests.test_ticket9848",
            "update.tests.AdvancedTests.test_update_annotated_multi_table_queryset",
            "update.tests.AdvancedTests.test_update_ordered_by_m2m_annotation",
            "update.tests.AdvancedTests.test_update_ordered_by_m2m_annotation_desc",
        },
        "QuerySet.extra() is not supported.": {
            "aggregation.tests.AggregateTestCase.test_exists_extra_where_with_aggregate",
            "annotations.tests.NonAggregateAnnotationTestCase.test_column_field_ordering",
            "annotations.tests.NonAggregateAnnotationTestCase.test_column_field_ordering_with_deferred",
            "basic.tests.ModelTest.test_extra_method_select_argument_with_dashes",
            "basic.tests.ModelTest.test_extra_method_select_argument_with_dashes_and_values",
            "defer.tests.DeferTests.test_defer_extra",
            "delete_regress.tests.Ticket19102Tests.test_ticket_19102_extra",
            "extra_regress.tests.ExtraRegressTests",
            "lookup.tests.LookupTests.test_values",
            "lookup.tests.LookupTests.test_values_list",
            "many_to_one.tests.ManyToOneTests.test_selects",
            "ordering.tests.OrderingTests.test_extra_ordering",
            "ordering.tests.OrderingTests.test_extra_ordering_quoting",
            "ordering.tests.OrderingTests.test_extra_ordering_with_table_name",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_union_multiple_models_with_values_list_and_order_by_extra_select",
            "queries.test_qs_combinators.QuerySetSetOperationTests.test_union_with_extra_and_values_list",
            "queries.tests.EscapingTests.test_ticket_7302",
            "queries.tests.Queries1Tests.test_tickets_1878_2939",
            "queries.tests.Queries1Tests.test_tickets_7087_12242",
            "queries.tests.Queries5Tests.test_extra_select_literal_percent_s",
            "queries.tests.Queries5Tests.test_ticket7256",
            "queries.tests.ValuesQuerysetTests.test_extra_multiple_select_params_values_order_by",
            "queries.tests.ValuesQuerysetTests.test_extra_select_params_values_order_in_extra",
            "queries.tests.ValuesQuerysetTests.test_extra_values",
            "queries.tests.ValuesQuerysetTests.test_extra_values_list",
            "queries.tests.ValuesQuerysetTests.test_extra_values_order_multiple",
            "queries.tests.ValuesQuerysetTests.test_extra_values_order_twice",
            "queries.tests.ValuesQuerysetTests.test_flat_extra_values_list",
            "queries.tests.ValuesQuerysetTests.test_named_values_list_with_fields",
            "queries.tests.ValuesQuerysetTests.test_named_values_list_without_fields",
            "select_related.tests.SelectRelatedTests.test_select_related_with_extra",
        },
        "Test inspects query for SQL": {
            "aggregation.tests.AggregateAnnotationPruningTests.test_non_aggregate_annotation_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_unreferenced_aggregate_annotation_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_unused_aliased_aggregate_pruned",
            "aggregation.tests.AggregateAnnotationPruningTests.test_referenced_aggregate_annotation_kept",
            "aggregation.tests.AggregateTestCase.test_count_star",
            "delete.tests.DeletionTests.test_only_referenced_fields_selected",
            "expressions.tests.ExistsTests.test_optimizations",
            "lookup.tests.LookupTests.test_in_ignore_none",
            "lookup.tests.LookupTests.test_lookup_direct_value_rhs_unwrapped",
            "lookup.tests.LookupTests.test_textfield_exact_null",
            "many_to_many.tests.ManyToManyQueryTests.test_count_join_optimization_disabled",
            "many_to_many.tests.ManyToManyQueryTests.test_exists_join_optimization_disabled",
            "many_to_many.tests.ManyToManyTests.test_custom_default_manager_exists_count",
            "migrations.test_commands.MigrateTests.test_migrate_syncdb_app_label",
            "migrations.test_commands.MigrateTests.test_migrate_syncdb_deferred_sql_executed_with_schemaeditor",
            "queries.tests.ExistsSql.test_exists",
            "queries.tests.Queries6Tests.test_col_alias_quoted",
            "schema.tests.SchemaTests.test_rename_column_renames_deferred_sql_references",
            "schema.tests.SchemaTests.test_rename_table_renames_deferred_sql_references",
        },
        "Test executes raw SQL.": {
            "aggregation.tests.AggregateTestCase.test_coalesced_empty_result_set",
            "aggregation_regress.tests.AggregationTests.test_annotate_with_extra",
            "aggregation_regress.tests.AggregationTests.test_annotation",
            "aggregation_regress.tests.AggregationTests.test_more_more3",
            "aggregation_regress.tests.AggregationTests.test_more_more_more3",
            "annotations.tests.NonAggregateAnnotationTestCase.test_raw_sql_with_inherited_field",
            "async.test_async_queryset.AsyncQuerySetTest.test_raw",
            "backends.base.test_base.ExecuteWrapperTests",
            "backends.tests.BackendTestCase.test_cursor_contextmanager",
            "backends.tests.BackendTestCase.test_cursor_executemany",
            "backends.tests.BackendTestCase.test_cursor_executemany_with_empty_params_list",
            "backends.tests.BackendTestCase.test_cursor_executemany_with_iterator",
            "backends.tests.BackendTestCase.test_duplicate_table_error",
            "backends.tests.BackendTestCase.test_queries",
            "backends.tests.BackendTestCase.test_queries_bare_where",
            "backends.tests.BackendTestCase.test_queries_limit",
            "backends.tests.BackendTestCase.test_queries_logger",
            "backends.tests.BackendTestCase.test_unicode_fetches",
            "backends.tests.EscapingChecks",
            "backends.test_utils.CursorWrapperTests",
            "custom_methods.tests.MethodsTests.test_custom_methods",
            "delete_regress.tests.DeleteLockingTest.test_concurrent_delete",
            "expressions.tests.BasicExpressionsTests.test_annotate_values_filter",
            "expressions.tests.BasicExpressionsTests.test_filtering_on_rawsql_that_is_boolean",
            "expressions.tests.BasicExpressionsTests.test_order_by_multiline_sql",
            "migrations.test_commands.MigrateTests.test_migrate_plan",
            "migrations.test_multidb.MultiDBOperationTests.test_run_sql_migrate_foo_router_with_hints",
            "migrations.test_operations.OperationTests.test_run_sql",
            "migrations.test_operations.OperationTests.test_run_sql_params",
            "migrations.test_operations.OperationTests.test_separate_database_and_state",
            "model_fields.test_jsonfield.TestQuerying.test_key_sql_injection_escape",
            "model_fields.test_jsonfield.TestQuerying.test_key_transform_raw_expression",
            "model_fields.test_jsonfield.TestQuerying.test_nested_key_transform_raw_expression",
            "multiple_database.tests.QueryTestCase.test_raw",
            "prefetch_related.tests.RawQuerySetTests",
            "queries.tests.Queries1Tests.test_order_by_rawsql",
            "raw_query.tests.RawQueryTests",
            "schema.test_logging.SchemaLoggerTests.test_extra_args",
            "schema.tests.SchemaTests.test_remove_constraints_capital_letters",
            "test_utils.tests.AllowedDatabaseQueriesTests.test_allowed_database_copy_queries",
            "timezones.tests.LegacyDatabaseTests.test_cursor_execute_accepts_naive_datetime",
            "timezones.tests.LegacyDatabaseTests.test_cursor_execute_returns_naive_datetime",
            "timezones.tests.LegacyDatabaseTests.test_raw_sql",
            "timezones.tests.NewDatabaseTests.test_cursor_execute_accepts_naive_datetime",
            "timezones.tests.NewDatabaseTests.test_cursor_execute_returns_naive_datetime",
            "timezones.tests.NewDatabaseTests.test_cursor_explicit_time_zone",
            "timezones.tests.NewDatabaseTests.test_raw_sql",
            "view_tests.tests.test_debug.DebugViewQueriesAllowedTests.test_handle_db_exception",
        },
        "Custom aggregations/functions with SQL don't work on MongoDB.": {
            "aggregation.tests.AggregateTestCase.test_add_implementation",
            "aggregation.tests.AggregateTestCase.test_multi_arg_aggregate",
            "aggregation.tests.AggregateTestCase.test_empty_result_optimization",
            "annotations.tests.NonAggregateAnnotationTestCase.test_custom_functions",
            "annotations.tests.NonAggregateAnnotationTestCase.test_custom_functions_can_ref_other_functions",
        },
        "Bilateral transform not implemented.": {
            "db_functions.tests.FunctionTests.test_func_transform_bilateral",
            "db_functions.tests.FunctionTests.test_func_transform_bilateral_multivalue",
        },
        "MongoDB does not support this database function.": {
            "db_functions.datetime.test_now.NowTests",
            "db_functions.math.test_sign.SignTests",
            "db_functions.text.test_chr.ChrTests",
            "db_functions.text.test_md5.MD5Tests",
            "db_functions.text.test_ord.OrdTests",
            "db_functions.text.test_pad.PadTests",
            "db_functions.text.test_repeat.RepeatTests",
            "db_functions.text.test_reverse.ReverseTests",
            "db_functions.text.test_right.RightTests",
            "db_functions.text.test_sha1.SHA1Tests",
            "db_functions.text.test_sha224.SHA224Tests",
            "db_functions.text.test_sha256.SHA256Tests",
            "db_functions.text.test_sha384.SHA384Tests",
            "db_functions.text.test_sha512.SHA512Tests",
        },
        "ExtractQuarter database function not supported.": {
            "db_functions.datetime.test_extract_trunc.DateFunctionTests.test_extract_quarter_func",
            "db_functions.datetime.test_extract_trunc.DateFunctionTests.test_extract_quarter_func_boundaries",
        },
        "TruncDate database function's tzinfo not supported.": {
            "model_fields.test_datetimefield.DateTimeFieldTests.test_lookup_date_with_use_tz",
            "timezones.tests.NewDatabaseTests.test_query_convert_timezones",
        },
        "MongoDB can't annotate ($project) a function like PI().": {
            "aggregation.tests.AggregateTestCase.test_aggregation_default_using_decimal_from_database",
            "db_functions.math.test_pi.PiTests.test",
        },
        "Can't cast from date to datetime without MongoDB interpreting the new value in UTC.": {
            "db_functions.comparison.test_cast.CastTests.test_cast_from_db_date_to_datetime",
            "db_functions.comparison.test_cast.CastTests.test_cast_from_db_datetime_to_time",
        },
        "Casting Python literals doesn't work.": {
            "db_functions.comparison.test_cast.CastTests.test_cast_from_python",
            "db_functions.comparison.test_cast.CastTests.test_cast_from_python_to_date",
            "db_functions.comparison.test_cast.CastTests.test_cast_from_python_to_datetime",
            "db_functions.comparison.test_cast.CastTests.test_cast_to_duration",
        },
        "inspectdb is not supported.": {
            "inspectdb.tests.InspectDBTestCase",
            "inspectdb.tests.InspectDBTransactionalTests",
        },
        "DatabaseIntrospection.get_table_description() not supported.": {
            "introspection.tests.IntrospectionTests.test_bigautofield",
            "introspection.tests.IntrospectionTests.test_get_table_description_col_lengths",
            "introspection.tests.IntrospectionTests.test_get_table_description_names",
            "introspection.tests.IntrospectionTests.test_get_table_description_nullable",
            "introspection.tests.IntrospectionTests.test_get_table_description_types",
            "introspection.tests.IntrospectionTests.test_smallautofield",
        },
        "MongoDB can't introspect primary key.": {
            "introspection.tests.IntrospectionTests.test_get_primary_key_column",
            "schema.tests.SchemaTests.test_alter_primary_key_the_same_name",
            "schema.tests.SchemaTests.test_primary_key",
        },
        "Known issue querying JSONField.": {
            # An ExpressionWrapper annotation with KeyTransform followed by
            # .filter(expr__isnull=False) doesn't use KeyTransformIsNull as it
            # needs to.
            "model_fields.test_jsonfield.TestQuerying.test_expression_wrapper_key_transform",
            # There is no way to distinguish between a JSON "null" (represented
            # by Value(None, JSONField())) and a SQL null (queried using the
            # isnull lookup). Both of these queries return both nulls.
            "model_fields.test_jsonfield.TestSaveLoad.test_json_null_different_from_sql_null",
            # Some queries with Q objects, e.g. Q(value__foo="bar"), don't work
            # properly, particularly with QuerySet.exclude().
            "model_fields.test_jsonfield.TestQuerying.test_lookup_exclude",
            "model_fields.test_jsonfield.TestQuerying.test_lookup_exclude_nonexistent_key",
            # Queries like like QuerySet.filter(value__j=None) incorrectly
            # returns objects where the key doesn't exist.
            "model_fields.test_jsonfield.TestQuerying.test_none_key",
            "model_fields.test_jsonfield.TestQuerying.test_none_key_exclude",
        },
        "Queries without a collection aren't supported on MongoDB.": {
            "queries.test_q.QCheckTests",
            "queries.test_query.TestQueryNoModel",
        },
        "MongoDB doesn't use CursorDebugWrapper.": {
            # It prevents RuntimeWarning for import-time queries from working.
            # https://github.com/django/django/commit/fbd16438f46bc2128926958ad24331da5d1b406f
            "apps.tests.QueryPerformingAppTests",
            "backends.tests.LastExecutedQueryTest.test_last_executed_query",
            "backends.tests.LastExecutedQueryTest.test_last_executed_query_with_duplicate_params",
            "backends.tests.LastExecutedQueryTest.test_query_encoding",
        },
        "Test not applicable for MongoDB's SQLCompiler.": {
            "queries.test_iterator.QuerySetIteratorTests",
        },
        "Support for views not implemented.": {
            "introspection.tests.IntrospectionTests.test_table_names_with_views",
        },
        "Connection health checks not implemented.": {
            "backends.base.test_base.ConnectionHealthChecksTests",
        },
        "transaction.atomic() is not supported.": {
            "backends.base.test_base.DatabaseWrapperLoggingTests",
            "migrations.test_executor.ExecutorTests.test_atomic_operation_in_non_atomic_migration",
            "migrations.test_operations.OperationTests.test_run_python_atomic",
        },
        "transaction.rollback() is not supported.": {
            "transactions.tests.AtomicMiscTests.test_mark_for_rollback_on_error_in_autocommit",
            "transactions.tests.AtomicMiscTests.test_mark_for_rollback_on_error_in_transaction",
            "transactions.tests.NonAutocommitTests.test_orm_query_after_error_and_rollback",
        },
        "migrate --fake-initial is not supported.": {
            "migrations.test_commands.MigrateTests.test_migrate_fake_initial",
            "migrations.test_commands.MigrateTests.test_migrate_fake_split_initial",
            "migrations.test_executor.ExecutorTests.test_soft_apply",
        },
        "Test works in isolation but fails on CI.": {
            # Probably something to do with lack of transaction support.
            "migration_test_data_persistence.tests.MigrationDataNormalPersistenceTestCase.test_persistence",
        },
        "Database caching not implemented.": {
            "cache.tests.CreateCacheTableForDBCacheTests",
            "cache.tests.DBCacheTests",
            "cache.tests.DBCacheWithTimeZoneTests",
        },
        "FilteredRelation not supported.": {
            # https://github.com/mongodb/django-mongodb-backend/issues/157
            "filtered_relation.tests.FilteredRelationAggregationTests",
            "filtered_relation.tests.FilteredRelationAnalyticalAggregationTests",
            "filtered_relation.tests.FilteredRelationTests",
            "queryset_pickle.tests.PickleabilityTestCase.test_pickle_filteredrelation",
        },
        "Broken test.": {
            # This test uses a database router that only allows the auth_user
            # table to be created on "other", but then writes users to
            # "default". It can raise IntegrityError when run alongside other
            # tests because the auth_user table isn't flushed between tests.
            "multiple_database.tests.AuthTestCase.test_dumpdata",
        },
        "ForeignObject is not supported.": {
            "foreign_object.test_agnostic_order_trimjoin.TestLookupQuery.test_deep_mixed_backward",
            "foreign_object.test_empty_join.RestrictedConditionsTests",
            "foreign_object.tests.MultiColumnFKTests",
            "foreign_object.tests.TestExtraJoinFilterQ",
        },
        "Custom lookups are not supported.": {
            "custom_lookups.tests.BilateralTransformTests",
            "custom_lookups.tests.LookupTests.test_basic_lookup",
            "custom_lookups.tests.LookupTests.test_custom_name_lookup",
            "custom_lookups.tests.LookupTests.test_div3_extract",
            "custom_lookups.tests.SubqueryTransformTests.test_subquery_usage",
        },
    }

    @cached_property
    def is_mongodb_6_3(self):
        return self.connection.get_database_version() >= (6, 3)



import copy
import time

import django
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.backends.utils import logger
from django.utils.functional import SimpleLazyObject
from django.utils.text import format_lazy
from django.utils.version import get_version_tuple
from pymongo.uri_parser import parse_uri as pymongo_parse_uri


def check_django_compatability():
    """
    Verify that this version of django-mongodb-backend is compatible with the
    installed version of Django. For example, any django-mongodb-backend 5.0.x is
    compatible with Django 5.0.y.
    """
    from . import __version__

    if django.VERSION[:2] != get_version_tuple(__version__)[:2]:
        A = django.VERSION[0]
        B = django.VERSION[1]
        raise ImproperlyConfigured(
            f"You must use the latest version of django-mongodb-backend {A}.{B}.x "
            f"with Django {A}.{B}.y (found django-mongodb-backend {__version__})."
        )


def parse_uri(uri, *, db_name=None, conn_max_age=0, test=None):
    """
    Convert the given uri into a dictionary suitable for Django's DATABASES
    setting.
    """
    uri = pymongo_parse_uri(uri)
    host = None
    port = None
    if uri["fqdn"]:
        # This is a SRV URI and the host is the fqdn.
        host = f"mongodb+srv://{uri['fqdn']}"
    else:
        nodelist = uri.get("nodelist")
        if len(nodelist) == 1:
            host, port = nodelist[0]
        elif len(nodelist) > 1:
            host = ",".join([f"{host}:{port}" for host, port in nodelist])
    db_name = db_name or uri["database"]
    if not db_name:
        raise ImproperlyConfigured("You must provide the db_name parameter.")
    settings_dict = {
        "ENGINE": "django_mongodb_backend",
        "NAME": db_name,
        "HOST": host,
        "PORT": port,
        "USER": uri.get("username"),
        "PASSWORD": uri.get("password"),
        "OPTIONS": uri.get("options"),
        "CONN_MAX_AGE": conn_max_age,
    }
    if "authSource" not in settings_dict["OPTIONS"] and uri["database"]:
        settings_dict["OPTIONS"]["authSource"] = uri["database"]
    if test:
        settings_dict["TEST"] = test
    return settings_dict


def prefix_validation_error(error, prefix, code, params):
    """
    Prefix a validation error message while maintaining the existing
    validation data structure.
    """
    if error.error_list == [error]:
        error_params = error.params or {}
        return ValidationError(
            # Messages can't simply be concatenated since they might require
            # their associated parameters to be expressed correctly which is
            # not something format_lazy() does. For example, proxied
            # ngettext calls require a count parameter and are converted
            # to an empty string if they are missing it.
            message=format_lazy(
                "{} {}",
                SimpleLazyObject(lambda: prefix % params),
                SimpleLazyObject(lambda: error.message % error_params),
            ),
            code=code,
            params={**error_params, **params},
        )
    return ValidationError(
        [prefix_validation_error(e, prefix, code, params) for e in error.error_list]
    )


def set_wrapped_methods(cls):
    """Initialize the wrapped methods on cls."""
    if hasattr(cls, "logging_wrapper"):
        for attr in cls.wrapped_methods:
            setattr(cls, attr, cls.logging_wrapper(attr))
        del cls.logging_wrapper
    return cls


@set_wrapped_methods
class OperationDebugWrapper:
    # The PyMongo database and collection methods that this backend uses.
    wrapped_methods = {
        "aggregate",
        "create_collection",
        "create_indexes",
        "drop",
        "index_information",
        "insert_many",
        "delete_many",
        "drop_index",
        "rename",
        "update_many",
    }

    def __init__(self, db, collection=None):
        self.collection = collection
        self.db = db
        use_collection = collection is not None
        self.collection_name = f"{collection.name}." if use_collection else ""
        self.wrapped = self.collection if use_collection else self.db.database

    def __getattr__(self, attr):
        return getattr(self.wrapped, attr)

    def profile_call(self, func, args=(), kwargs=None):
        start = time.monotonic()
        retval = func(*args, **kwargs or {})
        duration = time.monotonic() - start
        return duration, retval

    def log(self, op, duration, args, kwargs=None):
        # If kwargs are used by any operations in the future, they must be
        # added to this logging.
        msg = "(%.3f) %s"
        args = ", ".join(repr(arg) for arg in args)
        operation = f"db.{self.collection_name}{op}({args})"
        if len(settings.DATABASES) > 1:
            msg += f"; alias={self.db.alias}"
        self.db.queries_log.append(
            {
                "sql": operation,
                "time": "%.3f" % duration,
            }
        )
        logger.debug(
            msg,
            duration,
            operation,
            extra={
                "duration": duration,
                "sql": operation,
                "alias": self.db.alias,
            },
        )

    def logging_wrapper(method):
        def wrapper(self, *args, **kwargs):
            func = getattr(self.wrapped, method)
            # Collection.insert_many() mutates args (the documents) by adding
            #  _id. deepcopy() to avoid logging that version.
            original_args = copy.deepcopy(args)
            duration, retval = self.profile_call(func, args, kwargs)
            self.log(method, duration, original_args, kwargs)
            return retval

        return wrapper


@set_wrapped_methods
class OperationCollector(OperationDebugWrapper):
    def __init__(self, collected_sql=None, *, collection=None, db=None):
        super().__init__(db, collection)
        self.collected_sql = collected_sql

    def log(self, op, args, kwargs=None):
        args = ", ".join(repr(arg) for arg in args)
        operation = f"db.{self.collection_name}{op}({args})"
        self.collected_sql.append(operation)

    def logging_wrapper(method):
        def wrapper(self, *args, **kwargs):
            self.log(method, args, kwargs)

        return wrapper



from django.test import TestCase

from .models import Number


class NumericLookupTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.objs = Number.objects.bulk_create(Number(num=x) for x in range(5))
        # Null values should be excluded in less than queries.
        Number.objects.create()

    def test_lt(self):
        self.assertQuerySetEqual(Number.objects.filter(num__lt=3), self.objs[:3])

    def test_lte(self):
        self.assertQuerySetEqual(Number.objects.filter(num__lte=3), self.objs[:4])



from django.core.exceptions import ImproperlyConfigured
from django.db import connection
from django.db.backends.signals import connection_created
from django.test import SimpleTestCase, TestCase

from django_mongodb_backend.base import DatabaseWrapper


class DatabaseWrapperTests(SimpleTestCase):
    def test_database_name_empty(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        msg = 'settings.DATABASES is missing the "NAME" value.'
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            DatabaseWrapper(settings).get_connection_params()


class DatabaseWrapperConnectionTests(TestCase):
    def test_set_autocommit(self):
        self.assertIs(connection.get_autocommit(), True)
        connection.set_autocommit(False)
        self.assertIs(connection.get_autocommit(), False)
        connection.set_autocommit(True)
        self.assertIs(connection.get_autocommit(), True)

    def test_connection_created_database_attr(self):
        """
        connection.database is available in the connection_created signal.
        """
        data = {}

        def receiver(sender, connection, **kwargs):  # noqa: ARG001
            data["database"] = connection.database

        connection_created.connect(receiver)
        connection.close()
        # Accessing database implicitly connects.
        connection.database  # noqa: B018
        self.assertIs(data["database"], connection.database)
        connection.close()
        connection_created.disconnect(receiver)
        data.clear()
        connection.connect()
        self.assertEqual(data, {})



import datetime
import json
import re
import uuid
from decimal import Decimal

from bson.decimal128 import Decimal128
from django.conf import settings
from django.db import DataError
from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.models import TextField
from django.db.models.expressions import Combinable, Expression
from django.db.models.functions import Cast, Trunc
from django.utils import timezone
from django.utils.regex_helper import _lazy_re_compile


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "django_mongodb_backend.compiler"
    combine_operators = {
        Combinable.ADD: "add",
        Combinable.SUB: "subtract",
        Combinable.MUL: "multiply",
        Combinable.DIV: "divide",
        Combinable.POW: "pow",
        Combinable.MOD: "mod",
        Combinable.BITAND: "bitAnd",
        Combinable.BITOR: "bitOr",
        Combinable.BITXOR: "bitXor",
    }
    explain_options = {"comment", "verbosity"}

    def adapt_datefield_value(self, value):
        """Store DateField as datetime."""
        if value is None:
            return None
        return datetime.datetime.combine(value, datetime.datetime.min.time())

    def adapt_datetimefield_value(self, value):
        if value is None:
            return None
        # Expression values are adapted by the database.
        if hasattr(value, "resolve_expression"):
            return value
        if timezone.is_aware(value):
            if settings.USE_TZ:
                value = timezone.make_naive(value, self.connection.timezone)
            else:
                raise ValueError(
                    "MongoDB backend does not support timezone-aware "
                    "datetimes when USE_TZ is False."
                )
        return value

    def adapt_decimalfield_value(self, value, max_digits=None, decimal_places=None):
        """Store DecimalField as Decimal128."""
        if value is None:
            return None
        return Decimal128(value)

    def adapt_json_value(self, value, encoder):
        if encoder is None:
            return value
        try:
            return json.loads(json.dumps(value, cls=encoder))
        except json.decoder.JSONDecodeError as e:
            raise DataError from e

    def adapt_timefield_value(self, value):
        """Store TimeField as datetime."""
        if value is None:
            return None
        # Expression values are adapted by the database.
        if hasattr(value, "resolve_expression"):
            return value
        if timezone.is_aware(value):
            raise ValueError("MongoDB backend does not support timezone-aware times.")
        return datetime.datetime.combine(datetime.datetime.min.date(), value)

    def _get_arrayfield_converter(self, converter, *args, **kwargs):
        # Return a database converter that can be applied to a list of values.
        def convert_value(value, expression, connection):
            return [converter(x, expression, connection) for x in value]

        return convert_value

    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == "ArrayField":
            converters.extend(
                [
                    self._get_arrayfield_converter(converter)
                    for converter in self.get_db_converters(
                        Expression(output_field=expression.output_field.base_field)
                    )
                ]
            )
        elif internal_type == "DateField":
            # Trunc(... output_field="DateField") values must remain datetime
            # until Trunc.convert_value() so they can be converted from UTC
            # before truncation.
            if not isinstance(expression, Trunc):
                converters.append(self.convert_datefield_value)
        elif internal_type == "DateTimeField":
            if settings.USE_TZ:
                converters.append(self.convert_datetimefield_value)
        elif internal_type == "DecimalField":
            converters.append(self.convert_decimalfield_value)
        elif internal_type == "EmbeddedModelField":
            converters.append(self.convert_embeddedmodelfield_value)
        elif internal_type == "JSONField":
            converters.append(self.convert_jsonfield_value)
        elif internal_type == "TimeField":
            # Trunc(... output_field="TimeField") values must remain datetime
            # until Trunc.convert_value() so they can be converted from UTC
            # before truncation.
            if not isinstance(expression, Trunc):
                converters.append(self.convert_timefield_value)
        elif internal_type == "UUIDField":
            converters.append(self.convert_uuidfield_value)
        return converters

    def convert_datefield_value(self, value, expression, connection):
        if value is not None:
            value = value.date()
        return value

    def convert_datetimefield_value(self, value, expression, connection):
        if value is not None:
            value = timezone.make_aware(value, self.connection.timezone)
        return value

    def convert_decimalfield_value(self, value, expression, connection):
        if value is not None:
            # from Decimal128 to decimal.Decimal()
            try:
                value = value.to_decimal()
            except AttributeError:
                # `value` could be an integer in the case of an annotation
                # like ExpressionWrapper(Value(1), output_field=DecimalField().
                return Decimal(value)
        return value

    def convert_durationfield_value(self, value, expression, connection):
        if value is not None:
            try:
                value = datetime.timedelta(milliseconds=value)
            except TypeError:
                # `value` could be Decimal128 if doing a computation with
                # DurationField and Decimal128.
                value = datetime.timedelta(milliseconds=int(str(value)))
        return value

    def convert_embeddedmodelfield_value(self, value, expression, connection):
        if value is not None:
            # Apply database converters to each field of the embedded model.
            for field in expression.output_field.embedded_model._meta.fields:
                field_expr = Expression(output_field=field)
                converters = connection.ops.get_db_converters(
                    field_expr
                ) + field_expr.get_db_converters(connection)
                for converter in converters:
                    value[field.attname] = converter(value[field.attname], field_expr, connection)
        return value

    def convert_jsonfield_value(self, value, expression, connection):
        """
        Convert dict data to a string so that JSONField.from_db_value() can
        decode it using json.loads().
        """
        return json.dumps(value)

    def convert_timefield_value(self, value, expression, connection):
        if value is not None:
            value = value.time()
        return value

    def convert_uuidfield_value(self, value, expression, connection):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def combine_expression(self, connector, sub_expressions):
        lhs, rhs = sub_expressions
        if connector == Combinable.BITLEFTSHIFT:
            return {"$multiply": [lhs, {"$pow": [2, rhs]}]}
        if connector == Combinable.BITRIGHTSHIFT:
            return {"$floor": {"$divide": [lhs, {"$pow": [2, rhs]}]}}
        operator = self.combine_operators[connector]
        return {f"${operator}": sub_expressions}

    def prep_for_like_query(self, x):
        """Escape "x" for $regexMatch queries."""
        return re.escape(x)

    def quote_name(self, name):
        if name.startswith('"') and name.endswith('"'):
            return name  # Quoting once is enough.
        return name

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        """
        Return a list of the tables which will be passed as argument to
        execute_sql_flush().
        """
        return tables

    def execute_sql_flush(self, tables):
        for table in tables:
            if table.startswith("system."):
                # Do not drop system collections.
                continue

            collection = self.connection.database[table]
            options = collection.options()
            if not options.get("capped", False):
                collection.delete_many({})

    def prep_lookup_value(self, value, field, lookup):
        """
        Perform type-conversion on `value` before using as a filter parameter.
        """
        if getattr(field, "rel", None) is not None:
            field = field.rel.get_related_field()
        field_kind = field.get_internal_type()

        if lookup in ("in", "range"):
            return [
                self._prep_lookup_value(subvalue, field, field_kind, lookup) for subvalue in value
            ]
        return self._prep_lookup_value(value, field, field_kind, lookup)

    def _prep_lookup_value(self, value, field, field_kind, lookup):
        if value is None:
            return None

        if field_kind == "DecimalField":
            value = self.adapt_decimalfield_value(value, field.max_digits, field.decimal_places)
        return value

    def explain_query_prefix(self, format=None, **options):
        # Validate options.
        validated_options = {}
        if options:
            for valid_option in self.explain_options:
                value = options.pop(valid_option, None)
                if value is not None:
                    validated_options[valid_option] = value
        # super() raises an error if any options are left after the valid ones
        # are popped above.
        super().explain_query_prefix(format, **options)
        return validated_options

    def integer_field_range(self, internal_type):
        # MongODB doesn't enforce any integer constraints, but it supports
        # integers up to 64 bits.
        if internal_type in {
            "PositiveBigIntegerField",
            "PositiveIntegerField",
            "PositiveSmallIntegerField",
        }:
            return (0, 9223372036854775807)
        return (-9223372036854775808, 9223372036854775807)

    def prepare_join_on_clause(self, lhs_table, lhs_field, rhs_table, rhs_field):
        lhs_expr, rhs_expr = super().prepare_join_on_clause(
            lhs_table, lhs_field, rhs_table, rhs_field
        )
        # If the types are different, cast both to string.
        if lhs_field.db_type(self.connection) != rhs_field.db_type(self.connection):
            if lhs_field.db_type(self.connection) != "string":
                lhs_expr = Cast(lhs_expr, output_field=TextField())
            if rhs_field.db_type(self.connection) != "string":
                rhs_expr = Cast(rhs_expr, output_field=TextField())
        return lhs_expr, rhs_expr

    """Django uses these methods to generate SQL queries before it generates MQL queries."""

    # EXTRACT format cannot be passed in parameters.
    _extract_format_re = _lazy_re_compile(r"[A-Z_]+")

    def date_extract_sql(self, lookup_type, sql, params):
        if lookup_type == "week_day":
            # For consistency across backends, we return Sunday=1, Saturday=7.
            return f"EXTRACT(DOW FROM {sql}) + 1", params
        if lookup_type == "iso_week_day":
            return f"EXTRACT(ISODOW FROM {sql})", params
        if lookup_type == "iso_year":
            return f"EXTRACT(ISOYEAR FROM {sql})", params

        lookup_type = lookup_type.upper()
        if not self._extract_format_re.fullmatch(lookup_type):
            raise ValueError(f"Invalid lookup type: {lookup_type!r}")
        return f"EXTRACT({lookup_type} FROM {sql})", params

    def datetime_extract_sql(self, lookup_type, sql, params, tzname):
        if lookup_type == "second":
            # Truncate fractional seconds.
            return f"EXTRACT(SECOND FROM DATE_TRUNC(%s, {sql}))", ("second", *params)
        return self.date_extract_sql(lookup_type, sql, params)

    def datetime_trunc_sql(self, lookup_type, sql, params, tzname):
        return f"DATE_TRUNC(%s, {sql})", (lookup_type, *params)

    def date_trunc_sql(self, lookup_type, sql, params, tzname=None):
        return f"DATE_TRUNC(%s, {sql})", (lookup_type, *params)

    def datetime_cast_date_sql(self, sql, params, tzname):
        return f"({sql})::date", params

    def datetime_cast_time_sql(self, sql, params, tzname):
        return f"({sql})::time", params

    def format_for_duration_arithmetic(self, sql):
        return "INTERVAL %s MILLISECOND" % sql

    def time_trunc_sql(self, lookup_type, sql, params, tzname=None):
        return f"DATE_TRUNC(%s, {sql})::time", (lookup_type, *params)



from django.core.validators import BaseValidator, MaxLengthValidator, MinLengthValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import ngettext_lazy


class ArrayMaxLengthValidator(MaxLengthValidator):
    message = ngettext_lazy(
        "List contains %(show_value)d item, it should contain no more than %(limit_value)d.",
        "List contains %(show_value)d items, it should contain no more than %(limit_value)d.",
        "show_value",
    )


class ArrayMinLengthValidator(MinLengthValidator):
    message = ngettext_lazy(
        "List contains %(show_value)d item, it should contain no fewer than %(limit_value)d.",
        "List contains %(show_value)d items, it should contain no fewer than %(limit_value)d.",
        "show_value",
    )


@deconstructible
class LengthValidator(BaseValidator):
    message = ngettext_lazy(
        "List contains %(show_value)d item, it should contain %(limit_value)d.",
        "List contains %(show_value)d items, it should contain %(limit_value)d.",
        "show_value",
    )
    code = "length"

    def compare(self, a, b):
        return a != b

    def clean(self, x):
        return len(x)



from django.core.exceptions import FullResultSet
from django.db.models.aggregates import Aggregate
from django.db.models.expressions import Value


def is_direct_value(node):
    return not hasattr(node, "as_sql")


def process_lhs(node, compiler, connection):
    if not hasattr(node, "lhs"):
        # node is a Func or Expression, possibly with multiple source expressions.
        result = []
        for expr in node.get_source_expressions():
            if expr is None:
                continue
            try:
                result.append(expr.as_mql(compiler, connection))
            except FullResultSet:
                result.append(Value(True).as_mql(compiler, connection))
        if isinstance(node, Aggregate):
            return result[0]
        return result
    # node is a Transform with just one source expression, aliased as "lhs".
    if is_direct_value(node.lhs):
        return node
    return node.lhs.as_mql(compiler, connection)


def process_rhs(node, compiler, connection):
    rhs = node.rhs
    if hasattr(rhs, "as_mql"):
        if getattr(rhs, "subquery", False) and hasattr(node, "get_subquery_wrapping_pipeline"):
            value = rhs.as_mql(
                compiler, connection, get_wrapping_pipeline=node.get_subquery_wrapping_pipeline
            )
        else:
            value = rhs.as_mql(compiler, connection)
    else:
        _, value = node.process_rhs(compiler, connection)
        lookup_name = node.lookup_name
        # Undo Lookup.get_db_prep_lookup() putting params in a list.
        if lookup_name not in ("in", "range"):
            value = value[0]
    if hasattr(node, "prep_lookup_value_mongo"):
        value = node.prep_lookup_value_mongo(value)
    # No need to prepare expressions like F() objects.
    if hasattr(rhs, "resolve_expression"):
        return value
    return connection.ops.prep_lookup_value(value, node.lhs.output_field, node.lookup_name)


def regex_match(field, regex_vals, insensitive=False):
    regex = {"$concat": regex_vals} if isinstance(regex_vals, tuple) else regex_vals
    options = "i" if insensitive else ""
    return {"$regexMatch": {"input": {"$toString": field}, "regex": regex, "options": options}}



from .array import SimpleArrayField, SplitArrayField, SplitArrayWidget
from .embedded_model import EmbeddedModelField
from .objectid import ObjectIdField

__all__ = [
    "EmbeddedModelField",
    "SimpleArrayField",
    "SplitArrayField",
    "SplitArrayWidget",
    "ObjectIdField",
]



from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.models import Index
from pymongo import ASCENDING, DESCENDING


class DatabaseIntrospection(BaseDatabaseIntrospection):
    ORDER_DIR = {ASCENDING: "ASC", DESCENDING: "DESC"}

    def table_names(self, cursor=None, include_views=False):
        return sorted([x["name"] for x in self.connection.database.list_collections()])

    def get_constraints(self, cursor, table_name):
        indexes = self.connection.get_collection(table_name).index_information()
        constraints = {}
        for name, details in indexes.items():
            # Remove underscore prefix from "_id" columns in primary key index.
            if is_primary_key := name == "_id_":
                name = "id"
                details["key"] = [("id", 1)]
            constraints[name] = {
                "check": False,
                "columns": [field for field, order in details["key"]],
                "definition": None,
                "foreign_key": None,
                "index": True,
                "orders": [self.ORDER_DIR[order] for field, order in details["key"]],
                "primary_key": is_primary_key,
                "type": Index.suffix,
                "unique": details.get("unique", False),
                "options": {},
            }
        return constraints



from unittest.mock import patch

import pymongo
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase

from django_mongodb_backend import parse_uri


class ParseURITests(SimpleTestCase):
    def test_simple_uri(self):
        settings_dict = parse_uri("mongodb://cluster0.example.mongodb.net/myDatabase")
        self.assertEqual(settings_dict["ENGINE"], "django_mongodb_backend")
        self.assertEqual(settings_dict["NAME"], "myDatabase")
        self.assertEqual(settings_dict["HOST"], "cluster0.example.mongodb.net")
        self.assertEqual(settings_dict["OPTIONS"], {"authSource": "myDatabase"})

    def test_db_name(self):
        settings_dict = parse_uri("mongodb://cluster0.example.mongodb.net/", db_name="myDatabase")
        self.assertEqual(settings_dict["ENGINE"], "django_mongodb_backend")
        self.assertEqual(settings_dict["NAME"], "myDatabase")
        self.assertEqual(settings_dict["HOST"], "cluster0.example.mongodb.net")
        self.assertEqual(settings_dict["OPTIONS"], {})

    def test_db_name_overrides_default_auth_db(self):
        settings_dict = parse_uri(
            "mongodb://cluster0.example.mongodb.net/default_auth_db", db_name="myDatabase"
        )
        self.assertEqual(settings_dict["ENGINE"], "django_mongodb_backend")
        self.assertEqual(settings_dict["NAME"], "myDatabase")
        self.assertEqual(settings_dict["HOST"], "cluster0.example.mongodb.net")
        self.assertEqual(settings_dict["OPTIONS"], {"authSource": "default_auth_db"})

    def test_no_database(self):
        msg = "You must provide the db_name parameter."
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            parse_uri("mongodb://cluster0.example.mongodb.net")

    def test_srv_uri_with_options(self):
        uri = "mongodb+srv://my_user:my_password@cluster0.example.mongodb.net/my_database?retryWrites=true&w=majority"
        # patch() prevents a crash when PyMongo attempts to resolve the
        # nonexistent SRV record.
        with patch("dns.resolver.resolve"):
            settings_dict = parse_uri(uri)
        self.assertEqual(settings_dict["NAME"], "my_database")
        self.assertEqual(settings_dict["HOST"], "mongodb+srv://cluster0.example.mongodb.net")
        self.assertEqual(settings_dict["USER"], "my_user")
        self.assertEqual(settings_dict["PASSWORD"], "my_password")
        self.assertIsNone(settings_dict["PORT"])
        self.assertEqual(
            settings_dict["OPTIONS"],
            {"authSource": "my_database", "retryWrites": True, "w": "majority", "tls": True},
        )

    def test_localhost(self):
        settings_dict = parse_uri("mongodb://localhost/db")
        self.assertEqual(settings_dict["HOST"], "localhost")
        self.assertEqual(settings_dict["PORT"], 27017)

    def test_localhost_with_port(self):
        settings_dict = parse_uri("mongodb://localhost:27018/db")
        self.assertEqual(settings_dict["HOST"], "localhost")
        self.assertEqual(settings_dict["PORT"], 27018)

    def test_hosts_with_ports(self):
        settings_dict = parse_uri("mongodb://localhost:27017,localhost:27018/db")
        self.assertEqual(settings_dict["HOST"], "localhost:27017,localhost:27018")
        self.assertEqual(settings_dict["PORT"], None)

    def test_hosts_without_ports(self):
        settings_dict = parse_uri("mongodb://host1.net,host2.net/db")
        self.assertEqual(settings_dict["HOST"], "host1.net:27017,host2.net:27017")
        self.assertEqual(settings_dict["PORT"], None)

    def test_auth_source_in_query_string(self):
        settings_dict = parse_uri("mongodb://localhost/?authSource=auth", db_name="db")
        self.assertEqual(settings_dict["NAME"], "db")
        self.assertEqual(settings_dict["OPTIONS"], {"authSource": "auth"})

    def test_auth_source_in_query_string_overrides_defaultauthdb(self):
        settings_dict = parse_uri("mongodb://localhost/db?authSource=auth")
        self.assertEqual(settings_dict["NAME"], "db")
        self.assertEqual(settings_dict["OPTIONS"], {"authSource": "auth"})

    def test_conn_max_age(self):
        settings_dict = parse_uri("mongodb://localhost/db", conn_max_age=600)
        self.assertEqual(settings_dict["CONN_MAX_AGE"], 600)

    def test_test_kwarg(self):
        settings_dict = parse_uri("mongodb://localhost/db", test={"NAME": "test_db"})
        self.assertEqual(settings_dict["TEST"], {"NAME": "test_db"})

    def test_invalid_credentials(self):
        msg = "The empty string is not valid username"
        with self.assertRaisesMessage(pymongo.errors.InvalidURI, msg):
            parse_uri("mongodb://:@localhost")

    def test_no_scheme(self):
        with self.assertRaisesMessage(pymongo.errors.InvalidURI, "Invalid URI scheme"):
            parse_uri("cluster0.example.mongodb.net")
